/***********************************************************************************
* Fancy Tasks: Plasmoid providing a fancy representation of your tasks and launchers.
* Copyright (C) 2009 Michal Dutkiewicz aka Emdek <emdeck@gmail.com>
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*
***********************************************************************************/

#define PI 3.141592653

#include <config.h>

#include "FancyTasksApplet.h"
#include "FancyTasksIcon.h"
#include "FancyTasksJob.h"
#include "FancyTasksLauncher.h"
#include "FancyTasksSeparator.h"
#include "FancyTasksDropZone.h"

#include <cmath>

#include <QPainter>
#include <QDesktopWidget>
#include <QListWidgetItem>

#include <KRun>
#include <KMenu>
#include <KIcon>
#include <KLocale>
#include <KIconLoader>
#include <KWindowInfo>
#include <NETRootInfo>
#include <KFileDialog>
#include <KConfigDialog>
#include <KWindowSystem>

#include <Plasma/Theme>
#include <Plasma/Corona>
#include <Plasma/Animator>
#include <Plasma/Containment>

#ifdef FANCYTASKS_HAVE_COMPOSITING
#include <QX11Info>

#include <X11/extensions/Xrender.h>
#include <X11/extensions/Xcomposite.h>
#include <X11/extensions/shape.h>
#include <X11/Xlib.h>
#include <fixx11h.h>
#endif

K_EXPORT_PLASMA_APPLET(fancytasks, FancyTasksApplet)

FancyTasksApplet::FancyTasksApplet(QObject *parent, const QVariantList &args) : Plasma::Applet(parent, args)
{
    KGlobal::locale()->insertCatalog("fancytasks");

    setBackgroundHints(NoBackground);

    setHasConfigurationInterface(true);

    setAcceptDrops(true);

    m_animationTypes[NoAnimation] = QString();
    m_animationTypes[BounceAnimation] = "bounceAnimation";
    m_animationTypes[ZoomAnimation] = "zoomAnimation";
    m_animationTypes[BlinkAnimation] = "blinkAnimation";
    m_animationTypes[SpotlightAnimation] = "spotlightAnimation";
    m_animationTypes[RotateAnimation] = "rotateAnimation";
    m_animationTypes[JumpAnimation] = "jumpAnimation";
    m_animationTypes[FadeAnimation] = "fadeAnimation";
    m_animationTypes[GlowAnimation] = "glowAnimation";

    m_theme = new Plasma::FrameSvg(this);
    m_theme->setImagePath("widgets/fancytasks");
    m_theme->setEnabledBorders(Plasma::FrameSvg::AllBorders);

    m_background = m_theme;

    m_layout = new QGraphicsLinearLayout;
    m_layout->setContentsMargins(2, 2, 2, 2);
    m_layout->setSpacing(0);

    setLayout(m_layout);

    m_appletMaximumHeight = 100;

    m_initialFactor = 0;

    m_animation = 0;

    m_size = QSize(500, 100);

    resize(500, 100);
}

void FancyTasksApplet::init()
{
    KConfigGroup configuration = config();

    m_jobCloseMode = static_cast<CloseJobMode>(configuration.readEntry("jobCloseMode", static_cast<int>(DelayedClose)));
    m_moveAnimation = static_cast<AnimationType>(configuration.readEntry("moveAnimation", static_cast<int>(ZoomAnimation)));
    m_parabolicMoveAnimation = configuration.readEntry("parabolicMoveAnimation", true);
    m_demandsAttentionAnimation = static_cast<AnimationType>(configuration.readEntry("demandsAttentionAnimation", static_cast<int>(BlinkAnimation)));
    m_startupAnimation = static_cast<AnimationType>(configuration.readEntry("startupAnimation", static_cast<int>(BounceAnimation)));
    m_useThumbnails = configuration.readEntry("useThumbnails", false);
    m_titleLabelMode = static_cast<TitleLabelMode>(configuration.readEntry("titleLabelMode", static_cast<int>(NoLabel)));
    m_customBackgroundImage = configuration.readEntry("customBackgroundImage", QString());
    m_showOnlyCurrentDesktop = configuration.readEntry("showOnlyCurrentDesktop", false);
    m_showOnlyCurrentScreen = configuration.readEntry("showOnlyCurrentScreen", false);
    m_showOnlyMinimized = configuration.readEntry("showOnlyMinimized", false);
    m_connectJobsWithTasks = configuration.readEntry("connectJobsWithTasks", false);
    m_groupJobs = configuration.readEntry("groupJobs", true);
    m_groupingStrategy = static_cast<TaskManager::GroupManager::TaskGroupingStrategy>(configuration.readEntry("groupingStrategy", static_cast<int>(TaskManager::GroupManager::NoGrouping)));
    m_sortingStrategy = static_cast<TaskManager::GroupManager::TaskSortingStrategy>(configuration.readEntry("sortingStrategy", static_cast<int>(TaskManager::GroupManager::ManualSorting)));
    m_arrangement = configuration.readEntry("arrangement", QStringList("tasks"));
    m_initialFactor = ((m_moveAnimation == ZoomAnimation || m_moveAnimation == JumpAnimation)?0.7:0);

    if (!m_customBackgroundImage.isEmpty() && KUrl(m_customBackgroundImage).isValid())
    {
        m_background = new Plasma::FrameSvg(this);
        m_background->setImagePath(m_customBackgroundImage);
        m_background->setEnabledBorders(Plasma::FrameSvg::AllBorders);
    }

    m_groupManager = new TaskManager::GroupManager(this);
    m_groupManager->setGroupingStrategy(m_groupingStrategy);
    m_groupManager->setSortingStrategy(m_sortingStrategy);
    m_groupManager->setShowOnlyCurrentDesktop(m_showOnlyCurrentDesktop);
    m_groupManager->setShowOnlyCurrentScreen(m_showOnlyCurrentScreen);
    m_groupManager->setShowOnlyMinimized(m_showOnlyMinimized);

    if (containment())
    {
        m_groupManager->setScreen(containment()->screen());
    }

    connect(this, SIGNAL(activate()), this, SLOT(showMenu()));
    connect(m_groupManager->rootGroup(), SIGNAL(itemAdded(AbstractGroupableItem*)), this, SLOT(addTask(AbstractGroupableItem*)));
    connect(m_groupManager->rootGroup(), SIGNAL(itemRemoved(AbstractGroupableItem*)), this, SLOT(removeTask(AbstractGroupableItem*)));
    connect(m_groupManager->rootGroup(), SIGNAL(itemPositionChanged(AbstractGroupableItem*)), this, SLOT(changeTaskPosition(AbstractGroupableItem*)));

    m_groupManager->reconnect();

    m_dropZone = new FancyTasksDropZone(this);

    QGraphicsWidget *leftMargin = new QGraphicsWidget(this);
    leftMargin->setSizePolicy(QSizePolicy(QSizePolicy::Expanding, QSizePolicy::Expanding));

    QGraphicsWidget *rightMargin = new QGraphicsWidget(this);
    rightMargin->setSizePolicy(QSizePolicy(QSizePolicy::Expanding, QSizePolicy::Expanding));

    m_layout->addItem(leftMargin);
    m_layout->addItem(rightMargin);
    m_layout->addItem(m_dropZone);

    hideDropZone();

    int index = 1;

    for (int i = 0; i < m_arrangement.size(); ++i)
    {
        if (m_arrangement.at(i) == "separator")
        {
            if (i > 0 && !m_arrangement.at(i - 1).isEmpty())
            {
                FancyTasksSeparator *separator = new FancyTasksSeparator(m_theme, this);
                separator->setSize(m_itemSize);

                m_layout->insertItem(index, separator);

                ++index;
            }
        }
        else if (m_arrangement.at(i) != "tasks" && m_arrangement.at(i) != "jobs")
        {
            addLauncher(launcherForUrl(m_arrangement.at(i)), index);

            ++index;
        }
    }

    if (m_arrangement.contains("tasks"))
    {
        foreach (TaskManager::AbstractGroupableItem* abstractItem, m_groupManager->rootGroup()->members())
        {
            addTask(abstractItem);
        }
    }

    if (m_arrangement.contains("jobs"))
    {
        QStringList jobs = dataEngine("applicationjobs")->sources();

        for (int i = 0; i < jobs.count(); ++i)
        {
            addJob(jobs.at(i));
        }

        connect(dataEngine("applicationjobs"), SIGNAL(sourceAdded(const QString)), this, SLOT(addJob(const QString)));
        connect(dataEngine("applicationjobs"), SIGNAL(sourceRemoved(const QString)), this, SLOT(removeJob(const QString)));
    }
}

void FancyTasksApplet::paint(QPainter *painter, const QStyleOptionGraphicsItem *option, QWidget *widget)
{
    Q_UNUSED(option)
    Q_UNUSED(widget)

    painter->fillRect(rect(), Qt::transparent);

    if (formFactor() != Plasma::Horizontal && formFactor() != Plasma::Vertical)
    {
        painter->setRenderHint(QPainter::Antialiasing);
        painter->setCompositionMode(QPainter::CompositionMode_Source);

        m_background->resizeFrame(boundingRect().size());
        m_background->paintFrame(painter);
    }
}

void FancyTasksApplet::createConfigurationInterface(KConfigDialog *parent)
{
    KConfigGroup configuration = config();
    QWidget *generalWidget = new QWidget;
    QWidget *appearanceWidget = new QWidget;
    QWidget *arrangementWidget = new QWidget;

    m_generalUi.setupUi(generalWidget);
    m_appearanceUi.setupUi(appearanceWidget);
    m_arrangementUi.setupUi(arrangementWidget);

    m_generalUi.groupingStrategy->addItem(i18n("Do Not Group"), QVariant(TaskManager::GroupManager::NoGrouping));
    m_generalUi.groupingStrategy->addItem(i18n("Manually"), QVariant(TaskManager::GroupManager::ManualGrouping));
    m_generalUi.groupingStrategy->addItem(i18n("By Program Name"), QVariant(TaskManager::GroupManager::ProgramGrouping));
    m_generalUi.groupingStrategy->setCurrentIndex(m_generalUi.groupingStrategy->findData(QVariant(m_groupManager->groupingStrategy())));

    m_generalUi.sortingStrategy->addItem(i18n("Do Not Sort"), QVariant(TaskManager::GroupManager::NoSorting));
    m_generalUi.sortingStrategy->addItem(i18n("Manually"), QVariant(TaskManager::GroupManager::ManualSorting));
    m_generalUi.sortingStrategy->addItem(i18n("Alphabetically"), QVariant(TaskManager::GroupManager::AlphaSorting));
    m_generalUi.sortingStrategy->addItem(i18n("By Desktop"), QVariant(TaskManager::GroupManager::DesktopSorting));
    m_generalUi.sortingStrategy->setCurrentIndex(m_generalUi.sortingStrategy->findData(QVariant(m_groupManager->sortingStrategy())));

    m_generalUi.showOnlyCurrentDesktop->setChecked(m_showOnlyCurrentDesktop);
    m_generalUi.showOnlyCurrentScreen->setChecked(m_showOnlyCurrentScreen);
    m_generalUi.showOnlyMinimized->setChecked(m_showOnlyMinimized);
    m_generalUi.connectJobsWithTasks->setChecked(m_connectJobsWithTasks);
    m_generalUi.groupJobs->setChecked(m_groupJobs);

    m_generalUi.jobCloseMode->addItem(i18n("Instantly"), QVariant(InstantClose));
    m_generalUi.jobCloseMode->addItem(i18n("After delay"), QVariant(DelayedClose));
    m_generalUi.jobCloseMode->addItem(i18n("Manually"), QVariant(ManualClose));
    m_generalUi.jobCloseMode->setCurrentIndex(m_generalUi.jobCloseMode->findData(QVariant(m_jobCloseMode)));

    QStringList moveAnimationNames;
    moveAnimationNames.append(i18n("None"));
    moveAnimationNames.append(i18n("Zoom"));
    moveAnimationNames.append(i18n("Jump"));
    moveAnimationNames.append(i18n("Spotlight"));
    moveAnimationNames.append(i18n("Glow"));
    moveAnimationNames.append(i18n("Fade"));

    QList<AnimationType> moveAnimationIds;
    moveAnimationIds.append(NoAnimation);
    moveAnimationIds.append(ZoomAnimation);
    moveAnimationIds.append(JumpAnimation);
    moveAnimationIds.append(SpotlightAnimation);
    moveAnimationIds.append(GlowAnimation);
    moveAnimationIds.append(FadeAnimation);

    for (int i = 0; i < moveAnimationIds.size(); ++i)
    {
        m_appearanceUi.moveAnimation->addItem(moveAnimationNames.at(i), QVariant(moveAnimationIds.at(i)));
    }

    QStringList iconAnimationNames;
    iconAnimationNames.append(i18n("None"));
    iconAnimationNames.append(i18n("Bounce"));
    iconAnimationNames.append(i18n("Zoom"));
    iconAnimationNames.append(i18n("Blink"));
    iconAnimationNames.append(i18n("Spotlight"));
    iconAnimationNames.append(i18n("Rotate"));
    iconAnimationNames.append(i18n("Glow"));

    QList<AnimationType> iconAnimationIds;
    iconAnimationIds.append(NoAnimation);
    iconAnimationIds.append(BounceAnimation);
    iconAnimationIds.append(ZoomAnimation);
    iconAnimationIds.append(BlinkAnimation);
    iconAnimationIds.append(SpotlightAnimation);
    iconAnimationIds.append(RotateAnimation);
    iconAnimationIds.append(GlowAnimation);

    for (int i = 0; i < iconAnimationIds.size(); ++i)
    {
        m_appearanceUi.demandsAttentionAnimation->addItem(iconAnimationNames.at(i), QVariant(iconAnimationIds.at(i)));
        m_appearanceUi.startupAnimation->addItem(iconAnimationNames.at(i), QVariant(iconAnimationIds.at(i)));
    }

    m_appearanceUi.titleLabelMode->addItem(i18n("Never"), QVariant(NoLabel));
    m_appearanceUi.titleLabelMode->addItem(i18n("On mouse-over"), QVariant(LabelOnMouseOver));
    m_appearanceUi.titleLabelMode->addItem(i18n("Always"), QVariant(AlwaysShowLabel));
    m_appearanceUi.titleLabelMode->setCurrentIndex(m_appearanceUi.titleLabelMode->findData(QVariant(m_titleLabelMode)));

    m_appearanceUi.useThumbnails->setChecked(m_useThumbnails);
    m_appearanceUi.customBackgroundImage->setUrl(KUrl(m_customBackgroundImage));
    m_appearanceUi.customBackgroundImage->setFilter("image/svg+xml image/svg+xml-compressed");
    m_appearanceUi.moveAnimation->setCurrentIndex(moveAnimationIds.indexOf(m_moveAnimation));
    m_appearanceUi.parabolicMoveAnimation->setChecked(m_parabolicMoveAnimation);
    m_appearanceUi.demandsAttentionAnimation->setCurrentIndex(iconAnimationIds.indexOf(m_demandsAttentionAnimation));
    m_appearanceUi.startupAnimation->setCurrentIndex(iconAnimationIds.indexOf(m_startupAnimation));

    m_arrangementUi.removeButton->setIcon(KIcon("go-previous"));
    m_arrangementUi.addButton->setIcon(KIcon("go-next"));
    m_arrangementUi.moveUpButton->setIcon(KIcon("go-up"));
    m_arrangementUi.moveDownButton->setIcon(KIcon("go-down"));
    m_arrangementUi.availableActionsListWidget->addItem(i18n("--- separator ---"));


    if (!m_arrangement.contains("tasks"))
    {
        m_arrangementUi.availableActionsListWidget->addItem(i18n("--- tasks area ---"));
    }

    if (!m_arrangement.contains("jobs"))
    {
        m_arrangementUi.availableActionsListWidget->addItem(i18n("--- jobs area ---"));
    }

    for (int i = 0; i < m_arrangement.count(); ++i)
    {
        QListWidgetItem *item;

        if (m_arrangement.at(i) == "tasks")
        {
            item = new QListWidgetItem(i18n("--- tasks area ---"), m_arrangementUi.currentActionsListWidget);
        }
        else if (m_arrangement.at(i) == "jobs")
        {
            item = new QListWidgetItem(i18n("--- jobs area ---"), m_arrangementUi.currentActionsListWidget);
        }
        else if (m_arrangement.at(i) == "separator")
        {
            item = new QListWidgetItem(i18n("--- separator ---"), m_arrangementUi.currentActionsListWidget);
        }
        else
        {
            FancyTasksLauncher *launcher = launcherForUrl(KUrl(m_arrangement.at(i)));

            if (!launcher)
            {
                continue;
            }

            item = new QListWidgetItem(launcher->icon(), launcher->title(), m_arrangementUi.currentActionsListWidget);
            item->setToolTip(launcher->url().url());
        }

        m_arrangementUi.currentActionsListWidget->addItem(item);
    }

    moveAnimationTypeChanged(m_appearanceUi.moveAnimation->currentIndex());

    parent->addPage(generalWidget, i18n("General"), "go-home");
    parent->addPage(appearanceWidget, i18n("Appearance"), "preferences-desktop-theme");
    parent->addPage(arrangementWidget, i18n("Arrangement"), "format-list-unordered");

    connect(parent, SIGNAL(okClicked()), this, SLOT(configAccepted()));
    connect(m_appearanceUi.moveAnimation, SIGNAL(currentIndexChanged(int)), this, SLOT(moveAnimationTypeChanged(int)));
    connect(m_arrangementUi.availableActionsListWidget, SIGNAL(currentRowChanged(int)), this, SLOT(availableActionsCurrentItemChanged(int)));
    connect(m_arrangementUi.currentActionsListWidget, SIGNAL(currentRowChanged(int)), this, SLOT(currentActionsCurrentItemChanged(int)));
    connect(m_arrangementUi.removeButton, SIGNAL(clicked()), this, SLOT(removeItem()));
    connect(m_arrangementUi.addButton, SIGNAL(clicked()), this, SLOT(addItem()));
    connect(m_arrangementUi.moveUpButton, SIGNAL(clicked()), this, SLOT(moveUpItem()));
    connect(m_arrangementUi.moveDownButton, SIGNAL(clicked()), this, SLOT(moveDownItem()));
    connect(m_arrangementUi.addLauncherButton, SIGNAL(clicked()), this, SLOT(addLauncher()));
}

void FancyTasksApplet::configAccepted()
{
    KConfigGroup configuration = config();
    QStringList arrangement;

    if (!m_customBackgroundImage.isEmpty() && (m_appearanceUi.customBackgroundImage->url().path().isEmpty() || !m_appearanceUi.customBackgroundImage->url().isValid()))
    {
        m_background = m_theme;

        update();
    }
    else if (!m_appearanceUi.customBackgroundImage->url().path().isEmpty() && m_appearanceUi.customBackgroundImage->url().isValid())
    {
        if (m_background == m_theme)
        {
            m_background = new Plasma::FrameSvg(this);
            m_background->setEnabledBorders(Plasma::FrameSvg::AllBorders);
        }

        m_background->setImagePath(m_appearanceUi.customBackgroundImage->url().path());

        update();
    }

    for (int i = 0; i < m_arrangementUi.currentActionsListWidget->count(); ++i)
    {
        if (!m_arrangementUi.currentActionsListWidget->item(i)->toolTip().isEmpty())
        {
            arrangement.append(m_arrangementUi.currentActionsListWidget->item(i)->toolTip());
        }
        else if (m_arrangementUi.currentActionsListWidget->item(i)->text() == i18n("--- tasks area ---"))
        {
            arrangement.append("tasks");
        }
        else if (m_arrangementUi.currentActionsListWidget->item(i)->text() == i18n("--- jobs area ---"))
        {
            arrangement.append("jobs");
        }
        else
        {
            arrangement.append("separator");
        }
    }

    if (arrangement != m_arrangement)
    {
        if (m_arrangement.contains("tasks") && !arrangement.contains("tasks"))
        {
            QHash<TaskManager::AbstractGroupableItem*, QPointer<FancyTasksIcon> >::iterator iterator;

            for (iterator = m_taskIcons.begin(); iterator != m_taskIcons.end(); ++iterator)
            {
                m_layout->removeItem(iterator.value().data());

                iterator.value().data()->deleteLater();
            }
        }

        if (!m_arrangement.contains("jobs") && arrangement.contains("jobs"))
        {
            QStringList jobs = dataEngine("applicationjobs")->sources();

            for (int i = 0; i < jobs.count(); ++i)
            {
                addJob(jobs.at(i));
            }

            connect(dataEngine("applicationjobs"), SIGNAL(sourceAdded(const QString)), this, SLOT(addJob(const QString)));
            connect(dataEngine("applicationjobs"), SIGNAL(sourceRemoved(const QString)), this, SLOT(removeJob(const QString)));
        }
        else if (m_arrangement.contains("jobs") && !arrangement.contains("jobs"))
        {
            disconnect(dataEngine("applicationjobs"), SIGNAL(sourceAdded(const QString)), this, SLOT(addJob(const QString)));
            disconnect(dataEngine("applicationjobs"), SIGNAL(sourceRemoved(const QString)), this, SLOT(removeJob(const QString)));
        }

        for (int i = 0; i < m_layout->count(); ++i)
        {
            QObject *object = dynamic_cast<QObject*>(m_layout->itemAt(i));

            if (object->objectName() == "FancyTasksIcon" || object->objectName() == "FancyTasksSeparator")
            {
                if (object->objectName() == "FancyTasksIcon")
                {
                    QPointer<FancyTasksIcon> icon = dynamic_cast<FancyTasksIcon*>(m_layout->itemAt(i));

                    if (icon && icon->itemType() != FancyTasksIcon::Launcher)
                    {
                        continue;
                    }

                    m_launcherIcons.remove(icon->launcher());
                }

                m_layout->removeItem(m_layout->itemAt(i));

                delete object;

                --i;
            }
        }

        int index = 1;

        for (int i = 0; i < arrangement.size(); ++i)
        {
            if (arrangement.at(i) == "separator")
            {
                if (i > 0 && !arrangement.at(i - 1).isEmpty())
                {
                    m_layout->insertItem((index + ((m_arrangement.contains("tasks") && i >= m_arrangement.indexOf("tasks"))?m_taskIcons.count():0) + ((m_arrangement.contains("jobs") && i >= m_arrangement.indexOf("jobs"))?m_jobs.count():0)), new FancyTasksSeparator(m_theme, this));

                    ++index;
                }
            }
            else if (arrangement.at(i) != "tasks" && arrangement.at(i) != "jobs")
            {
                addLauncher(launcherForUrl(arrangement.at(i)), index);

                ++index;
            }
        }

        if (!m_arrangement.contains("tasks") && arrangement.contains("tasks"))
        {
            foreach (TaskManager::AbstractGroupableItem* abstractItem, m_groupManager->rootGroup()->members())
            {
                addTask(abstractItem);
            }
        }
    }

    m_moveAnimation = static_cast<AnimationType>(m_appearanceUi.moveAnimation->itemData(m_appearanceUi.moveAnimation->currentIndex()).toInt());
    m_parabolicMoveAnimation = m_appearanceUi.parabolicMoveAnimation->isChecked();
    m_demandsAttentionAnimation = static_cast<AnimationType>(m_appearanceUi.demandsAttentionAnimation->itemData(m_appearanceUi.demandsAttentionAnimation->currentIndex()).toInt());
    m_startupAnimation = static_cast<AnimationType>(m_appearanceUi.startupAnimation->itemData(m_appearanceUi.startupAnimation->currentIndex()).toInt());
    m_useThumbnails = m_appearanceUi.useThumbnails->isChecked();
    m_titleLabelMode = static_cast<TitleLabelMode>(m_appearanceUi.titleLabelMode->itemData(m_appearanceUi.titleLabelMode->currentIndex()).toInt());
    m_customBackgroundImage = (m_appearanceUi.customBackgroundImage->url().isValid()?m_appearanceUi.customBackgroundImage->url().path():QString());
    m_showOnlyCurrentDesktop = m_generalUi.showOnlyCurrentDesktop->isChecked();
    m_showOnlyCurrentScreen = m_generalUi.showOnlyCurrentScreen->isChecked();
    m_showOnlyMinimized = m_generalUi.showOnlyMinimized->isChecked();
    m_connectJobsWithTasks = m_generalUi.connectJobsWithTasks->isChecked();
    m_groupJobs = m_generalUi.groupJobs->isChecked();
    m_groupingStrategy = static_cast<TaskManager::GroupManager::TaskGroupingStrategy>(m_generalUi.groupingStrategy->itemData(m_generalUi.groupingStrategy->currentIndex()).toInt());
    m_sortingStrategy = static_cast<TaskManager::GroupManager::TaskSortingStrategy>(m_generalUi.sortingStrategy->itemData(m_generalUi.sortingStrategy->currentIndex()).toInt());
    m_jobCloseMode = static_cast<CloseJobMode>(m_generalUi.jobCloseMode->itemData(m_generalUi.jobCloseMode->currentIndex()).toInt());
    m_arrangement = arrangement;
    m_initialFactor = ((m_moveAnimation == ZoomAnimation || m_moveAnimation == JumpAnimation)?0.7:0);

    m_groupManager->setGroupingStrategy(m_groupingStrategy);
    m_groupManager->setSortingStrategy(m_sortingStrategy);
    m_groupManager->setShowOnlyCurrentDesktop(m_showOnlyCurrentDesktop);
    m_groupManager->setShowOnlyCurrentScreen(m_showOnlyCurrentScreen);
    m_groupManager->setShowOnlyMinimized(m_showOnlyMinimized);

    if (containment())
    {
        m_groupManager->setScreen(containment()->screen());
    }

    m_groupManager->reconnect();

    updateSize();

    configuration.writeEntry("moveAnimation", static_cast<int>(m_moveAnimation));
    configuration.writeEntry("parabolicMoveAnimation", m_parabolicMoveAnimation);
    configuration.writeEntry("demandsAttentionAnimation", static_cast<int>(m_demandsAttentionAnimation));
    configuration.writeEntry("startupAnimation", static_cast<int>(m_startupAnimation));
    configuration.writeEntry("useThumbnails", m_useThumbnails);
    configuration.writeEntry("titleLabelMode", static_cast<int>(m_titleLabelMode));
    configuration.writeEntry("customBackgroundImage", m_customBackgroundImage);
    configuration.writeEntry("showOnlyCurrentDesktop", m_showOnlyCurrentDesktop);
    configuration.writeEntry("showOnlyCurrentScreen", m_showOnlyCurrentScreen);
    configuration.writeEntry("showOnlyMinimized", m_showOnlyMinimized);
    configuration.writeEntry("connectJobsWithTasks", m_connectJobsWithTasks);
    configuration.writeEntry("groupJobs", m_groupJobs);
    configuration.writeEntry("groupingStrategy", static_cast<int>(m_groupingStrategy));
    configuration.writeEntry("sortingStrategy", static_cast<int>(m_sortingStrategy));
    configuration.writeEntry("jobCloseMode", static_cast<int>(m_jobCloseMode));
    configuration.writeEntry("arrangement", m_arrangement);

    emit configNeedsSaving();
}

void FancyTasksApplet::moveAnimationTypeChanged(int option)
{
    m_appearanceUi.parabolicMoveAnimation->setEnabled(option != 0 && m_appearanceUi.moveAnimation->itemData(option).toInt() != static_cast<int>(JumpAnimation));
}

void FancyTasksApplet::availableActionsCurrentItemChanged(int row)
{
    m_arrangementUi.addButton->setEnabled(row >= 0);

    if (row >= 0)
    {
        m_arrangementUi.descriptionLabel->setText(m_arrangementUi.availableActionsListWidget->currentItem()->text());
    }
}

void FancyTasksApplet::currentActionsCurrentItemChanged(int row)
{
    m_arrangementUi.removeButton->setEnabled(row >= 0);
    m_arrangementUi.moveUpButton->setEnabled(row > 0);
    m_arrangementUi.moveDownButton->setEnabled(row >= 0 && row < (m_arrangementUi.currentActionsListWidget->count() - 1));

    if (row >= 0)
    {
        m_arrangementUi.descriptionLabel->setText(m_arrangementUi.currentActionsListWidget->currentItem()->text());
    }
}

void FancyTasksApplet::removeItem()
{
    if (m_arrangementUi.currentActionsListWidget->currentRow() >= 0)
    {
        QListWidgetItem *currentItem = m_arrangementUi.currentActionsListWidget->takeItem(m_arrangementUi.currentActionsListWidget->currentRow());

        if (currentItem->text() != i18n("--- separator ---"))
        {
            m_arrangementUi.availableActionsListWidget->addItem(currentItem);
        }
        else
        {
            delete currentItem;
        }

        m_arrangementUi.currentActionsListWidget->setCurrentItem(NULL);
        m_arrangementUi.availableActionsListWidget->setCurrentItem(NULL);
    }
}

void FancyTasksApplet::addItem()
{
    if (m_arrangementUi.availableActionsListWidget->currentRow() >= 0)
    {
        QListWidgetItem *currentItem = m_arrangementUi.availableActionsListWidget->takeItem(m_arrangementUi.availableActionsListWidget->currentRow());

        if (currentItem->text() == i18n("--- separator ---"))
        {
            m_arrangementUi.availableActionsListWidget->insertItem(0, currentItem->clone());
        }

        m_arrangementUi.currentActionsListWidget->insertItem((m_arrangementUi.currentActionsListWidget->currentRow() + 1), currentItem);
        m_arrangementUi.currentActionsListWidget->setCurrentItem(currentItem);
        m_arrangementUi.availableActionsListWidget->setCurrentItem(NULL);
    }
}

void FancyTasksApplet::moveUpItem()
{
    const int currentRow = m_arrangementUi.currentActionsListWidget->currentRow();

    if (currentRow > 0)
    {
        QListWidgetItem *currentItem = m_arrangementUi.currentActionsListWidget->takeItem(currentRow);

        m_arrangementUi.currentActionsListWidget->insertItem((currentRow - 1), currentItem);
        m_arrangementUi.currentActionsListWidget->setCurrentItem(currentItem);
    }
}

void FancyTasksApplet::moveDownItem()
{
    const int currentRow = m_arrangementUi.currentActionsListWidget->currentRow();

    if (currentRow < (m_arrangementUi.currentActionsListWidget->count() - 1))
    {
        QListWidgetItem *currentItem = m_arrangementUi.currentActionsListWidget->takeItem(currentRow);

        m_arrangementUi.currentActionsListWidget->insertItem((currentRow + 1), currentItem);

        m_arrangementUi.currentActionsListWidget->setCurrentItem(currentItem);
    }
}

void FancyTasksApplet::addLauncher()
{
    KFileDialog *dialog = new KFileDialog(KUrl("~"), "", NULL);
    dialog->setWindowModality(Qt::NonModal);
    dialog->setMode(KFile::File | KFile::Directory);
    dialog->setOperationMode(KFileDialog::Opening);
    dialog->exec();

    if (!dialog->selectedUrl().isEmpty())
    {
        FancyTasksLauncher* launcher = new FancyTasksLauncher(dialog->selectedUrl(), this);

        QListWidgetItem *item = new QListWidgetItem(launcher->icon(), launcher->title(), m_arrangementUi.currentActionsListWidget);
        item->setToolTip(launcher->url().pathOrUrl());

        m_arrangementUi.currentActionsListWidget->addItem(item);

        delete launcher;
    }

    delete dialog;
}

void FancyTasksApplet::constraintsEvent(Plasma::Constraints constraints)
{
    setBackgroundHints(NoBackground);

    if (constraints & Plasma::LocationConstraint)
    {
        m_layout->setOrientation((formFactor() == Plasma::Vertical)?Qt::Vertical:Qt::Horizontal);
    }
}

void FancyTasksApplet::resizeEvent(QGraphicsSceneResizeEvent *event)
{
    if (formFactor() == Plasma::Vertical)
    {
        m_appletMaximumHeight = event->newSize().width();
    }
    else
    {
        m_appletMaximumHeight = event->newSize().height();
    }

    updateSize();
}

void FancyTasksApplet::dragMoveEvent(QGraphicsSceneDragDropEvent *event)
{
    if (immutability() == Plasma::Mutable && (event->mimeData()->hasFormat("FancyTasksLauncher") || KUrl::List::canDecode(event->mimeData())))
    {
        int index = 0;

        if (formFactor() == Plasma::Vertical)
        {
            if (event->pos().y() > (boundingRect().height() / 2))
            {
                index = (m_layout->count() - 1);
            }
        }
        else if (event->pos().x() > (boundingRect().width() / 2))
        {
            index = (m_layout->count() - 1);
        }

        if (m_dropZone->index() != index || !m_dropZone->isVisible())
        {
            m_dropZone->show(index);
        }
    }
}

void FancyTasksApplet::dragLeaveEvent(QGraphicsSceneDragDropEvent *event)
{
    Q_UNUSED(event)

    QTimer::singleShot(500, m_dropZone, SLOT(hide()));
}

void FancyTasksApplet::dropEvent(QGraphicsSceneDragDropEvent *event)
{
    if (!KUrl::List::canDecode(event->mimeData()))
    {
        event->ignore();

        return;
    }

    KUrl::List droppedUrls = KUrl::List::fromMimeData(event->mimeData());

    for (int i = 0; i < droppedUrls.count(); ++i)
    {
        new KRun(droppedUrls[i], NULL);
    }
}

void FancyTasksApplet::hoverLeaveEvent(QGraphicsSceneHoverEvent *event)
{
    Q_UNUSED(event)

    if (m_animation)
    {
        Plasma::Animator::self()->stopCustomAnimation(m_animation);
    }

    m_animation = Plasma::Animator::self()->customAnimation(5, 10, Plasma::Animator::LinearCurve, this, "moveAnimation");

    m_activeItem = -1;
}

void FancyTasksApplet::wheelEvent(QGraphicsSceneWheelEvent *event)
{
    event->accept();

    int steps = (event->delta() / 120);

    m_activeItem += steps;

    while (m_activeItem < 0 || m_activeItem >= (m_layout->count() - 1))
    {
        if (m_activeItem < 0)
        {
            m_activeItem += (m_layout->count() - 1);
        }
        else
        {
            m_activeItem -= (m_layout->count() - 1);
        }
    }

    moveAnimation(1);

    if (event->modifiers() & Qt::ShiftModifier)
    {
        return;
    }

    FancyTasksIcon *icon = static_cast<FancyTasksIcon*>(m_layout->itemAt(m_activeItem + 1));

    if (icon)
    {
        icon->activateWindow(true);
    }
}

void FancyTasksApplet::addTask(AbstractGroupableItem* abstractItem)
{
    if (!m_arrangement.contains("tasks") || !abstractItem)
    {
        return;
    }

    QPointer<FancyTasksIcon> icon = NULL;
    TaskManager::TaskItem *task = NULL;

    if (abstractItem->isGroupItem() && m_groupManager->groupingStrategy() != TaskManager::GroupManager::ManualGrouping)
    {
        TaskManager::TaskGroup *group = static_cast<TaskManager::TaskGroup*>(abstractItem);

        if (group->name().isEmpty())
        {
            group->setName(static_cast<TaskManager::TaskItem*>(group->members().at(0))->task()->visibleName());
        }
    }
    else if (!abstractItem->isGroupItem())
    {
        task = static_cast<TaskManager::TaskItem*>(abstractItem);
    }

    QHash<FancyTasksIcon*, QDateTime>::iterator iterator;

    if (!abstractItem->name().isEmpty())
    {
        for (iterator = m_removedStartups.begin(); iterator != m_removedStartups.end(); ++iterator)
        {
            if (abstractItem->name().contains(iterator.key()->title(), Qt::CaseInsensitive) && iterator.key()->itemType() == FancyTasksIcon::Startup)
            {
                icon = iterator.key();
                icon->setTask(abstractItem);

                m_removedStartups.erase(iterator);

                break;
            }
        }
    }

    if (!icon)
    {
        int index = (m_groupManager->rootGroup()->members().indexOf(abstractItem) + m_arrangement.indexOf("tasks") + 1);

        if (m_arrangement.contains("jobs") && index >= m_arrangement.indexOf("jobs"))
        {
            index += m_jobs.count();
        }

        icon = new FancyTasksIcon(abstractItem, launcherForTask(task), NULL, this);
        icon->setSize(m_itemSize);

        m_layout->insertItem(index, icon);

        updateSize();
    }

    m_taskIcons[abstractItem] = icon;
}

void FancyTasksApplet::removeTask(AbstractGroupableItem* abstractItem)
{
    if (!m_taskIcons.contains(abstractItem))
    {
        return;
    }

    QPointer<FancyTasksIcon> icon = m_taskIcons[abstractItem];

    if (icon->itemType() == FancyTasksIcon::Startup)
    {
        m_removedStartups[icon] = QDateTime::currentDateTime();

        QTimer::singleShot(2000, this, SLOT(cleanupRemovedStartups()));

        return;
    }

    m_taskIcons.remove(abstractItem);

    if (!icon)
    {
       return;
    }

    m_layout->removeItem(icon);

    icon->deleteLater();
}

void FancyTasksApplet::changeTaskPosition(AbstractGroupableItem* abstractItem)
{
    if (!m_arrangement.contains("tasks"))
    {
        return;
    }

    FancyTasksIcon* icon = m_taskIcons[abstractItem];

    if (!icon)
    {
        return;
    }

    TaskManager::AbstractGroupableItem *temp = icon->abstractItem();

    int index = (m_groupManager->rootGroup()->members().indexOf(temp) + m_arrangement.indexOf("tasks") + 1);

    if ((m_layout->itemAt(index)) == icon)
    {
        return;
    }

    m_layout->removeItem(icon);
    m_layout->insertItem(index, icon);
}

void FancyTasksApplet::addLauncher(FancyTasksLauncher *launcher, int index)
{
    if (!launcher || m_launcherIcons.contains(launcher))
    {
        return;
    }

    if (m_arrangement.contains("tasks") && index >= m_arrangement.indexOf("tasks"))
    {
        index += m_taskIcons.count();
    }

    if (m_arrangement.contains("jobs") && index >= m_arrangement.indexOf("jobs"))
    {
        index += m_jobs.count();
    }

    FancyTasksIcon *icon = new FancyTasksIcon(NULL, launcher, NULL, this);
    icon->setSize(m_itemSize);

    m_launcherIcons[launcher] = icon;

    m_layout->insertItem(index, icon);

    updateSize();
}

void FancyTasksApplet::removeLauncher(FancyTasksLauncher *launcher)
{
    if (!m_launcherIcons.contains(launcher))
    {
        return;
    }

    if (launcher)
    {
        m_launchers.removeAll(launcher);
    }

    FancyTasksIcon *icon = m_launcherIcons[launcher];

    m_launcherIcons.remove(launcher);

    if (!icon)
    {
       return;
    }

    m_layout->removeItem(icon);

    icon->deleteLater();
}

void FancyTasksApplet::addJob(const QString &source)
{
    if (!m_arrangement.contains("jobs") || m_jobs.contains(source))
    {
        return;
    }

    FancyTasksJob *job = new FancyTasksJob(source, this);

    m_jobsQueue.enqueue(job);
    m_jobs[source] = job;

    QTimer::singleShot(1500, this, SLOT(showJob()));
}

void FancyTasksApplet::removeJob(const QString &source, bool force)
{
    QPointer<FancyTasksJob> job = m_jobs[source];

    if (!job)
    {
        m_jobIcons.remove(m_jobs[source]);
        m_jobs.remove(source);

        return;
    }

    if (!force && !dataEngine("applicationjobs")->sources().contains(source))
    {
        job->setFinished(true);
    }

    if (job->closeOnFinish() && !job->hasError())
    {
        force = true;
    }
    else if (!force && m_jobCloseMode == DelayedClose && job->closeOnFinish())
    {
        QTimer::singleShot(5000, job, SLOT(close()));
    }

    if (!force)
    {
        return;
    }

    m_jobIcons.remove(m_jobs[source]);
    m_jobs.remove(source);

    if (!job)
    {
       return;
    }

    job->destroy();
}

void FancyTasksApplet::showJob()
{
    QPointer<FancyTasksJob> job = m_jobsQueue.dequeue();

    if (!job)
    {
        return;
    }

    if (job->isFinished())
    {
        job->close();

        return;
    }

    QString className;

    if (m_jobCloseMode != ManualClose)
    {
        job->setCloseOnFinish(true);
    }

    if (m_connectJobsWithTasks)
    {
        QHash<TaskManager::AbstractGroupableItem*, QPointer<FancyTasksIcon> >::iterator tasksIterator;

        for (tasksIterator = m_taskIcons.begin(); tasksIterator != m_taskIcons.end(); ++tasksIterator)
        {
            if (!tasksIterator.value())
            {
                continue;
            }

            if (tasksIterator.key()->isGroupItem() && m_groupManager->groupingStrategy() != TaskManager::GroupManager::ManualGrouping)
            {
                TaskManager::TaskGroup *group = static_cast<TaskManager::TaskGroup*>(tasksIterator.key());

                if (group->name().isEmpty())
                {
                    className = KWindowSystem::windowInfo(static_cast<TaskManager::TaskItem*>(group->members().at(0))->task()->window(), 0, NET::WM2WindowClass).windowClassName();
                }
            }
            else if (!tasksIterator.key()->isGroupItem())
            {
                className = KWindowSystem::windowInfo(static_cast<TaskManager::TaskItem*>(tasksIterator.key())->task()->window(), 0, NET::WM2WindowClass).windowClassName();
            }

            if (job->application().contains(className, Qt::CaseInsensitive))
            {
                tasksIterator.value()->addJob(job);

                return;
            }
        }
    }

    if (m_groupJobs)
    {
        QHash<FancyTasksJob*, QPointer<FancyTasksIcon> >::iterator jobsIterator;

        for (jobsIterator = m_jobIcons.begin(); jobsIterator != m_jobIcons.end(); ++jobsIterator)
        {
            if (!jobsIterator.value())
            {
                continue;
            }

            if (job->application() == jobsIterator.key()->application())
            {
                jobsIterator.value()->addJob(job);

                return;
            }
        }

    }

    FancyTasksIcon *icon = new FancyTasksIcon(NULL, NULL, job, this);
    icon->setSize(m_itemSize);

    int index = (m_arrangement.indexOf("jobs") + m_jobs.count());

    if (m_arrangement.contains("tasks") && index >= m_arrangement.indexOf("tasks"))
    {
        index += m_taskIcons.count();
    }

    m_layout->insertItem((m_taskIcons.count() + m_arrangement.indexOf("tasks") + 1), icon);

    updateSize();

    m_jobIcons[job] = icon;
}

void FancyTasksApplet::cleanupRemovedStartups()
{
    QMutableHashIterator<FancyTasksIcon*, QDateTime> iterator(m_removedStartups);

    while (iterator.hasNext())
    {
        iterator.next();

        if (iterator.value().secsTo(QDateTime::currentDateTime()) > 1 && iterator.key()->itemType() == FancyTasksIcon::Startup)
        {
            iterator.key()->deleteLater();

            iterator.remove();
        }
    }
}

void FancyTasksApplet::itemHoverMoved(QObject *item, qreal across)
{
    int index = 0;

    m_activeItem = 0;

    if (item)
    {
        for (int i = 0; i < m_layout->count(); ++i)
        {
            QObject *object = dynamic_cast<QObject*>(m_layout->itemAt(i));

            if (object->objectName() != "FancyTasksIcon" && object->objectName() != "FancyTasksSeparator")
            {
                continue;
            }

            if (object == item)
            {
                m_activeItem = index;

                break;
            }

            if (m_layout->itemAt(i)->geometry().width() > 1 || object->objectName() == "FancyTasksSeparator")
            {
                ++index;
            }
        }
    }

    m_across = across;

    moveAnimation(1);
}

void FancyTasksApplet::itemHoverLeft()
{
    m_activeItem = -1;

    if (m_animation)
    {
        Plasma::Animator::self()->stopCustomAnimation(m_animation);
    }

    m_animation = Plasma::Animator::self()->customAnimation(5, 10, Plasma::Animator::LinearCurve, this, "moveAnimation");
}

void FancyTasksApplet::moveAnimation(qreal progress)
{
    int index = 0;
    qreal factor;

    for (int i = 0; i < m_layout->count(); ++i)
    {
        QObject *object = dynamic_cast<QObject*>(m_layout->itemAt(i));

        if (object->objectName() != "FancyTasksIcon")
        {
            if (object->objectName() == "FancyTasksSeparator")
            {
                ++index;
            }

            continue;
        }

        FancyTasksIcon *icon = static_cast<FancyTasksIcon*>(m_layout->itemAt(i));

        if (!icon || !icon->isVisible())
        {
            continue;
        }

        if (index == m_activeItem)
        {
            factor = 1;
        }
        else if (!m_parabolicMoveAnimation || m_activeItem < 0 || index < (m_activeItem - 3) || index > (m_activeItem + 3))
        {
            factor = m_initialFactor;
        }
        else if (index < m_activeItem)
        {
            factor = (m_initialFactor + ((1 - m_initialFactor) * (cos(((index - m_activeItem - m_across + 1) / 3) * PI) + 1)) / 2);
        }
        else
        {
            factor = (m_initialFactor + ((1 - m_initialFactor) * (cos(((index - m_activeItem - m_across) / 3) * PI) + 1)) / 2);
        }

        if (icon->factor() != factor)
        {
            icon->setFactor(icon->factor() + ((factor - icon->factor()) * progress));
        }

        ++index;
    }
}

void FancyTasksApplet::needsVisualFocus()
{
    m_lastAttentionDemand = QDateTime::currentDateTime();

    emit activate();
}

void FancyTasksApplet::showMenu()
{
    bool previousIconLauncher = false;
    bool previousIconTask = false;

    if (m_lastAttentionDemand.isValid() && m_lastAttentionDemand.secsTo(QDateTime::currentDateTime()) < 1)
    {
        return;
    }

    KMenu *menu = new KMenu;

    for (int i = 0; i < m_layout->count(); ++i)
    {
        if (dynamic_cast<QObject*>(m_layout->itemAt(i))->objectName() != "FancyTasksIcon")
        {
            continue;
        }

        FancyTasksIcon *icon = static_cast<FancyTasksIcon*>(m_layout->itemAt(i));

        if (!icon || !icon->isVisible())
        {
            continue;
        }

        if ((previousIconLauncher && icon->itemType() != FancyTasksIcon::Launcher) || (previousIconTask && icon->itemType() == FancyTasksIcon::Launcher))
        {
            menu->addSeparator();
        }

        QAction *action = menu->addAction(icon->icon(), icon->title());
        QFont font = QFont(action->font());

        if (icon->itemType() == FancyTasksIcon::Launcher)
        {
            font.setItalic(true);

            previousIconLauncher = true;
            previousIconTask = false;
        }
        else if (icon->demandsAttention())
        {
            font.setBold(true);
        }

        if (icon->itemType() == FancyTasksIcon::Group)
        {
            action->setMenu(new TaskManager::GroupPopupMenu(NULL, icon->group(), m_groupManager));
        }
        else
        {
            connect(action, SIGNAL(triggered()), icon, SLOT(activate()));
        }

        if (icon->itemType() != FancyTasksIcon::Launcher)
        {
            previousIconLauncher = false;
            previousIconTask = true;
        }

        icon->blockToolTip();

        action->setFont(font);
    }

    menu->exec(QCursor::pos());

    delete menu;
}

void FancyTasksApplet::showDropZone(int index)
{
    if (m_dropZone->index() != index)
    {
        m_layout->removeItem(m_dropZone);
        m_layout->insertItem(index, m_dropZone);

        m_dropZone->show(index);
    }
}

void FancyTasksApplet::hideDropZone()
{
    if (!m_dropZone->isUnderMouse())
    {
        m_layout->removeItem(m_dropZone);
        m_layout->insertItem((m_layout->count() - 1), m_dropZone);

        m_dropZone->hide();
    }
}

void FancyTasksApplet::updateSize()
{
    QSize size;
    int separatorsGap = -1;

    m_itemSize = m_appletMaximumHeight;

    qreal itemSize = (m_itemSize * 0.8);

    if (location() == Plasma::Floating)
    {
        m_appletMaximumWidth = (m_background->marginSize(Plasma::LeftMargin) + m_background->marginSize(Plasma::RightMargin));
    }
    else
    {
        m_appletMaximumWidth = 0;
    }

    for (int i = 0; i < m_layout->count(); ++i)
    {
        QString objectName = dynamic_cast<QObject*>(m_layout->itemAt(i))->objectName();

        if (objectName == "FancyTasksDropZone")
        {
            if (m_dropZone->isVisible())
            {
                m_appletMaximumWidth += (itemSize / 2);
            }

            continue;
        }

        if (objectName == "FancyTasksSeparator")
        {
            FancyTasksSeparator *separator = dynamic_cast<FancyTasksSeparator*>(m_layout->itemAt(i));

            if (separatorsGap == 0)
            {
                if (separator->isVisible())
                {
                    separator->hide();
                }
            }
            else
            {
                if (!separator->isVisible())
                {
                    separator->show();
                }

                m_appletMaximumWidth += (itemSize / 4);

                separatorsGap = 0;
            }

            continue;
        }

        if (objectName == "FancyTasksIcon")
        {
            FancyTasksIcon *icon = static_cast<FancyTasksIcon*>(m_layout->itemAt(i));

            if (!icon || !icon->isVisible())
            {
                continue;
            }

            if (separatorsGap >= 0)
            {
                ++separatorsGap;
            }

            if (m_moveAnimation == JumpAnimation || (m_moveAnimation == ZoomAnimation && (!i || (m_parabolicMoveAnimation && i > 4))))
            {
                m_appletMaximumWidth += (m_initialFactor * itemSize);
            }
            else
            {
                m_appletMaximumWidth += itemSize;
            }
        }
    }

    if (formFactor() == Plasma::Vertical)
    {
        size = QSize(m_appletMaximumHeight, m_appletMaximumWidth);

        setPreferredHeight(m_appletMaximumHeight);

        setMinimumHeight(m_appletMaximumHeight);
    }
    else
    {
        size = QSize(m_appletMaximumWidth, m_appletMaximumHeight);

        setPreferredWidth(m_appletMaximumWidth);

        setMinimumWidth(m_appletMaximumWidth);
    }

    if (size == m_size)
    {
        return;
    }

    m_size = size;

    if (location() != Plasma::Floating)
    {
        emit sizeHintChanged(Qt::PreferredSize);

        emit sizeHintChanged(Qt::MinimumSize);
    }

    resize(size);

    update();

    emit sizeChanged(m_itemSize);
}

void FancyTasksApplet::urlChanged(const KUrl &oldUrl, KUrl &newUrl)
{
    KConfigGroup configuration = config();

    if (newUrl.url().isEmpty())
    {
        if (m_arrangement.contains(oldUrl.url()))
        {
            m_arrangement.removeAll(oldUrl.url());

            removeLauncher(launcherForUrl(oldUrl));
        }
    }
    else if (oldUrl.url().isEmpty())
    {
        if (!m_arrangement.contains(newUrl.url()))
        {
            m_arrangement.append(newUrl.url());

            addLauncher(launcherForUrl(newUrl), 0);
        }
    }
    else if (m_arrangement.contains(oldUrl.url()))
    {
        m_arrangement.replace(m_arrangement.indexOf(oldUrl.url()), newUrl.url());

        for (int i = 0; i < m_launchers.count(); ++i)
        {
            if (m_launchers.at(i)->url() == oldUrl)
            {
                m_launchers.at(i)->setUrl(newUrl);

                break;
            }
        }
    }

    configuration.writeEntry("arrangement", m_arrangement);

    emit configNeedsSaving();
}

void FancyTasksApplet::itemDropped(FancyTasksIcon *icon, int index)
{
    if (icon->itemType() == FancyTasksIcon::Task || icon->itemType() == FancyTasksIcon::Group)
    {
        index -= (m_arrangement.indexOf("tasks") + 1);

        if (index < 0)
        {
            index = 0;
        }

        m_groupManager->manualSortingRequest(icon->abstractItem(), index);

        return;
    }

    if (immutability() != Plasma::Mutable)
    {
        return;
    }

    KConfigGroup configuration = config();
    QStringList arrangement;

    if (m_arrangement.contains("tasks") && index >= m_arrangement.indexOf("tasks") && index <= (m_arrangement.indexOf("tasks") + m_taskIcons.count()))
    {
        index = 1;
    }

    m_layout->removeItem(icon);
    m_layout->insertItem(index, icon);

    for (int i = 0; i < m_layout->count(); ++i)
    {
        QString objectName = dynamic_cast<QObject*>(m_layout->itemAt(i))->objectName();

        if (objectName == "FancyTasksIcon")
        {
            FancyTasksIcon *icon = static_cast<FancyTasksIcon*>(m_layout->itemAt(i));

            if (icon && icon->itemType() == FancyTasksIcon::Launcher)
            {
                arrangement.append(icon->launcher()->url().pathOrUrl());
            }
            else if (!arrangement.contains("tasks"))
            {
                arrangement.append("tasks");
            }
        }
        else if (objectName == "FancyTasksSeparator")
        {
            arrangement.append("separator");
        }
    }

    m_arrangement = arrangement;

    configuration.writeEntry("arrangement", m_arrangement);

    emit configNeedsSaving();
}

void FancyTasksApplet::itemDragged(FancyTasksIcon *icon, QPointF position, const QMimeData *mimeData)
{
    if (((mimeData->hasFormat("FancyTasksTask") || mimeData->hasFormat("FancyTasksGroup")) && m_groupManager->sortingStrategy() == TaskManager::GroupManager::ManualSorting) || (immutability() == Plasma::Mutable && (mimeData->hasFormat("FancyTasksLauncher") || KUrl::List::canDecode(mimeData))))
    {
        int index = 0;

        for (int i = 0; i < m_layout->count(); ++i)
        {
            QObject *object = dynamic_cast<QObject*>(m_layout->itemAt(i));

            if (object->objectName() != "FancyTasksIcon" && object->objectName() != "FancyTasksSeparator")
            {
                continue;
            }

            if (!object)
            {
                continue;
            }

            if (object == icon)
            {
                break;
            }

            ++index;
        }

        if (formFactor() == Plasma::Vertical)
        {
            if (position.y() > (icon->boundingRect().height() / 2))
            {
                ++index;
            }
        }
        else if (position.x() > (icon->boundingRect().width() / 2))
        {
            ++index;
        }

        if (mimeData->hasFormat("FancyTasksTask") || mimeData->hasFormat("FancyTasksGroup"))
        {
            if (index < m_arrangement.indexOf("tasks"))
            {
                index = m_arrangement.indexOf("tasks");
            }
        }

        else if (m_arrangement.contains("tasks") && index >= m_arrangement.indexOf("tasks") && index <= (m_arrangement.indexOf("tasks") + m_taskIcons.count()))
        {
            index = 0;
        }

        showDropZone(index + 1);
    }
}

FancyTasksLauncher* FancyTasksApplet::launcherForUrl(KUrl url)
{
    FancyTasksLauncher *launcher = NULL;

    if (!url.isValid())
    {
        return launcher;
    }

    for (int i = 0; i < m_launchers.count(); ++i)
    {
        if (m_launchers.at(i)->url() == url)
        {
            launcher = m_launchers.at(i);

            break;
        }
    }

    if (!launcher)
    {
        launcher = new FancyTasksLauncher(url, this);

        m_launchers.append(launcher);
    }

    return launcher;
}

FancyTasksLauncher* FancyTasksApplet::launcherForTask(TaskManager::TaskItem *task)
{
        if (!task || task->task().isNull())
        {
            return (NULL);
        }

        QString className;
        WId window = task->task()->window();

        if (window)
        {
            className = KWindowSystem::windowInfo(window, 0, NET::WM2WindowClass).windowClassName();
        }

        for (int i = 0; i < m_launchers.count(); ++i)
        {
            if (!m_launchers.at(i)->isExecutable())
            {
                continue;
            }

            if ((!className.isEmpty() && (m_launchers.at(i)->title().contains(className, Qt::CaseInsensitive) || m_launchers.at(i)->exec().contains(className, Qt::CaseInsensitive))) || task->name().contains(m_launchers.at(i)->title(), Qt::CaseInsensitive))
            {
                return (m_launchers.at(i));
            }
        }

    return (NULL);
}

TaskManager::GroupManager* FancyTasksApplet::groupManager()
{
    return m_groupManager;
}

Plasma::Svg* FancyTasksApplet::theme()
{
    return m_theme;
}

QHash<FancyTasksApplet::AnimationType, QString> FancyTasksApplet::animationTypes()
{
    return m_animationTypes;
}

QStringList FancyTasksApplet::arrangement() const
{
    return m_arrangement;
}

FancyTasksApplet::TitleLabelMode FancyTasksApplet::titleLabelMode() const
{
    return m_titleLabelMode;
}

FancyTasksApplet::AnimationType FancyTasksApplet::moveAnimation() const
{
    return m_moveAnimation;
}

FancyTasksApplet::AnimationType FancyTasksApplet::demandsAttentionAnimation() const
{
    return m_demandsAttentionAnimation;
}

FancyTasksApplet::AnimationType FancyTasksApplet::startupAnimation() const
{
    return m_startupAnimation;
}

qreal FancyTasksApplet::initialFactor() const
{
    return m_initialFactor;
}

qreal FancyTasksApplet::itemSize() const
{
    return m_itemSize;
}

bool FancyTasksApplet::parabolicMoveAnimation() const
{
    return m_parabolicMoveAnimation;
}

bool FancyTasksApplet::useThumbnails() const
{
    return m_useThumbnails;
}

QPoint FancyTasksApplet::popupPosition(QSize size, QRect geometry, bool center) const
{
    QPoint position(geometry.left(), geometry.top());

    switch (location())
    {
        case Plasma::BottomEdge:
            position = QPoint((position.x() - (center?((size.width() - geometry.width()) / 2):0)), (position.y() - size.height()));
        break;
        case Plasma::TopEdge:
            position = QPoint((position.x() - (center?((size.width() - geometry.width()) / 2):0)), (position.y() + geometry.height()));
        break;
        case Plasma::LeftEdge:
            position = QPoint((position.x() + geometry.width()), (position.y() - (center?((size.height() - geometry.height()) / 2):0)));
        break;
        case Plasma::RightEdge:
            position = QPoint((position.x() - size.width()), (position.y() - (center?((size.height() - geometry.height()) / 2):0)));
        break;
        default:
            if (position.y() - size.height() > 0)
            {
                position = QPoint((position.x() - (center?((size.width() - geometry.width()) / 2):0)), (position.y() - size.height()));
            }
            else
            {
                position = QPoint((position.x() - (center?((size.width() - geometry.width()) / 2):0)), (position.y() + geometry.height()));
            }
    }

    QRect screenRect = containment()->corona()->screenGeometry(containment()?containment()->screen():-1);

    if (location() != Plasma::LeftEdge && position.rx() + size.width() > screenRect.right())
    {
        position.rx() -= ((position.rx() + size.width()) - screenRect.right());
    }

    if (location() != Plasma::TopEdge && position.ry() + size.height() > screenRect.bottom())
    {
        position.ry() -= ((position.ry() + size.height()) - screenRect.bottom());
    }

    position.rx() = qMax(0, position.rx());

    return position;
}

QPixmap FancyTasksApplet::windowPreview(WId window, int size)
{
    QPixmap thumbnail;

#ifdef FANCYTASKS_HAVE_COMPOSITING
    if (KWindowSystem::compositingActive())
    {
        Display *display = QX11Info::display();
        XWindowAttributes attributes;

        XCompositeRedirectWindow(display, window, CompositeRedirectAutomatic);

        XGetWindowAttributes(display, window, &attributes);

        XRenderPictFormat *format = XRenderFindVisualFormat(display, attributes.visual);

        if (format)
        {
            bool hasAlpha  = (format->type == PictTypeDirect && format->direct.alphaMask);
            int x = attributes.x;
            int y = attributes.y;
            int width = attributes.width;
            int height = attributes.height;

            XRenderPictureAttributes pictureAttributes;
            pictureAttributes.subwindow_mode = IncludeInferiors;

            Picture picture = XRenderCreatePicture(display, window, format, CPSubwindowMode, &pictureAttributes);

            XserverRegion region = XFixesCreateRegionFromWindow(display, window, WindowRegionBounding);

            XFixesTranslateRegion(display, region, -x, -y);
            XFixesSetPictureClipRegion(display, picture, 0, 0, region);
            XFixesDestroyRegion(display, region);

            XShapeSelectInput(display, window, ShapeNotifyMask);

            thumbnail = QPixmap(width, height);
            thumbnail.fill(Qt::transparent);

            XRenderComposite(display, (hasAlpha?PictOpOver:PictOpSrc), picture, None, thumbnail.x11PictureHandle(), 0, 0, 0, 0, 0, 0, width, height);
        }
    }
#endif

    if (thumbnail.isNull() && KWindowSystem::windowInfo(window, NET::XAWMState).mappingState() == NET::Visible)
    {
        thumbnail = QPixmap::grabWindow(QApplication::desktop()->winId()).copy(KWindowSystem::windowInfo(window, NET::WMGeometry | NET::WMFrameExtents).frameGeometry());
    }

    if (!thumbnail.isNull())
    {
        if (thumbnail.width() > thumbnail.height())
        {
            thumbnail = thumbnail.scaledToWidth(size, Qt::SmoothTransformation);
        }
        else
        {
            thumbnail = thumbnail.scaledToHeight(size, Qt::SmoothTransformation);
        }
    }

    return thumbnail;
}

#include "FancyTasksApplet.moc"
