/***********************************************************************************
* Fancy Tasks: Plasmoid providing a fancy representation of your tasks and launchers.
* Copyright (C) 2009 Michal Dutkiewicz aka Emdek <emdeck@gmail.com>
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*
***********************************************************************************/

#ifndef FANCYTASKSAPPLET_HEADER
#define FANCYTASKSAPPLET_HEADER

#include <QHash>
#include <QQueue>
#include <QTimer>
#include <QPointer>
#include <QMimeData>
#include <QDateTime>
#include <QGraphicsLinearLayout>
#include <QGraphicsSceneResizeEvent>
#include <QGraphicsSceneDragDropEvent>

#include <KUrl>

#include <taskmanager/task.h>
#include <taskmanager/taskitem.h>
#include <taskmanager/taskactions.h>
#include <taskmanager/taskmanager.h>
#include <taskmanager/groupmanager.h>

#include <Plasma/Applet>
#include <Plasma/FrameSvg>
#include <Plasma/DataEngine>

#include "ui_general.h"
#include "ui_appearance.h"
#include "ui_arrangement.h"

class FancyTasksIcon;
class FancyTasksJob;
class FancyTasksLauncher;
class FancyTasksDropZone;

using TaskManager::GroupPtr;
using TaskManager::TaskPtr;
using TaskManager::StartupPtr;
using TaskManager::AbstractGroupableItem;
using TaskManager::GroupManager;
using TaskManager::TaskItem;
using TaskManager::TaskGroup;

class FancyTasksApplet : public Plasma::Applet
{
    Q_OBJECT

    public:
        FancyTasksApplet(QObject *parent, const QVariantList &args);

        enum TitleLabelMode { NoLabel = 0, LabelOnMouseOver, AlwaysShowLabel };
        enum CloseJobMode { InstantClose, DelayedClose, ManualClose };
        enum AnimationType { NoAnimation = 0, ZoomAnimation, JumpAnimation, RotateAnimation, BounceAnimation, BlinkAnimation, GlowAnimation, SpotlightAnimation, FadeAnimation };

        void init();
        void dropEvent(QGraphicsSceneDragDropEvent *event);
        void itemDropped(FancyTasksIcon *icon, int index);
        void itemDragged(FancyTasksIcon *icon, QPointF position, const QMimeData *mimeData);
        FancyTasksLauncher* launcherForUrl(KUrl url);
        FancyTasksLauncher* launcherForTask(TaskManager::TaskItem *task);
        TaskManager::GroupManager* groupManager();
        Plasma::Svg* theme();
        QHash<AnimationType, QString> animationTypes();
        QStringList arrangement() const;
        TitleLabelMode titleLabelMode() const;
        AnimationType moveAnimation() const;
        AnimationType demandsAttentionAnimation() const;
        AnimationType startupAnimation() const;
        bool parabolicMoveAnimation() const;
        bool useThumbnails() const;
        qreal initialFactor() const;
        qreal itemSize() const;
        QPoint popupPosition(QSize size, QRect geometry, bool center = false) const;
        static QPixmap windowPreview(WId window, int size);

    protected:
        void paint(QPainter *painter, const QStyleOptionGraphicsItem *option, QWidget *widget);
        void createConfigurationInterface(KConfigDialog *parent);
        void resizeEvent(QGraphicsSceneResizeEvent *event);
        void dragMoveEvent(QGraphicsSceneDragDropEvent *event);
        void dragLeaveEvent(QGraphicsSceneDragDropEvent *event);
        void hoverLeaveEvent(QGraphicsSceneHoverEvent *event);
        void wheelEvent(QGraphicsSceneWheelEvent *event);

    public slots:
        void constraintsEvent(Plasma::Constraints constraints);
        void configAccepted();
        void moveAnimationTypeChanged(int option);
        void availableActionsCurrentItemChanged(int row);
        void currentActionsCurrentItemChanged(int row);
        void removeItem();
        void addItem();
        void moveUpItem();
        void moveDownItem();
        void addLauncher();
        void addTask(AbstractGroupableItem* abstractItem);
        void removeTask(AbstractGroupableItem* abstractItem);
        void changeTaskPosition(AbstractGroupableItem* abstractItem);
        void addLauncher(FancyTasksLauncher *launcher, int index);
        void removeLauncher(FancyTasksLauncher *launcher);
        void addJob(const QString &source);
        void removeJob(const QString &source, bool force = false);
        void showJob();
        void cleanupRemovedStartups();
        void itemHoverMoved(QObject *item, qreal across);
        void itemHoverLeft();
        void moveAnimation(qreal progress);
        void needsVisualFocus();
        void showMenu();
        void showDropZone(int index);
        void hideDropZone();
        void updateSize();
        void urlChanged(const KUrl &oldUrl, KUrl &newUrl);

    private:
        QGraphicsLinearLayout *m_layout;
        GroupManager *m_groupManager;
        QQueue<QPointer<FancyTasksJob> > m_jobsQueue;
        QList<QPointer<FancyTasksLauncher> > m_launchers;
        QHash<QString, QPointer<FancyTasksJob> > m_jobs;
        QHash<FancyTasksLauncher*, QPointer<FancyTasksIcon> > m_launcherIcons;
        QHash<FancyTasksJob*, QPointer<FancyTasksIcon> > m_jobIcons;
        QHash<AbstractGroupableItem*, QPointer<FancyTasksIcon> > m_taskIcons;
        QHash<FancyTasksIcon*, QDateTime> m_removedStartups;
        QHash<AnimationType, QString> m_animationTypes;
        QDateTime m_lastAttentionDemand;
        QSize m_size;
        Plasma::FrameSvg *m_theme;
        Plasma::FrameSvg *m_background;
        FancyTasksDropZone *m_dropZone;
        TaskManager::GroupManager::TaskGroupingStrategy m_groupingStrategy;
        TaskManager::GroupManager::TaskSortingStrategy m_sortingStrategy;
        QString m_customBackgroundImage;
        QStringList m_arrangement;
        TitleLabelMode m_titleLabelMode;
        CloseJobMode m_jobCloseMode;
        AnimationType m_moveAnimation;
        AnimationType m_demandsAttentionAnimation;
        AnimationType m_startupAnimation;
        qreal m_appletMaximumWidth;
        qreal m_appletMaximumHeight;
        qreal m_initialFactor;
        qreal m_across;
        qreal m_itemSize;
        int m_activeItem;
        int m_animation;
        bool m_parabolicMoveAnimation;
        bool m_useThumbnails;
        bool m_showOnlyCurrentDesktop;
        bool m_showOnlyCurrentScreen;
        bool m_showOnlyMinimized;
        bool m_connectJobsWithTasks;
        bool m_groupJobs;
        Ui::general m_generalUi;
        Ui::appearance m_appearanceUi;
        Ui::arrangement m_arrangementUi;

    signals:
        void sizeChanged(qreal size);
};

#endif
