/***********************************************************************************
* Fancy Tasks: Plasmoid providing a fancy representation of your tasks and launchers.
* Copyright (C) 2009 Michal Dutkiewicz aka Emdek <emdeck@gmail.com>
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*
***********************************************************************************/

#define PI 3.141592653

#include "FancyTasksIcon.h"
#include "FancyTasksLauncher.h"
#include "FancyTasksJob.h"
#include "FancyTasksToolTip.h"

#include <cmath>

#include <QMimeData>
#include <QTransform>
#include <QApplication>
#include <QFontMetrics>
#include <QGraphicsView>
#include <QVarLengthArray>

#include <KMenu>
#include <KIcon>
#include <KLocale>
#include <KIconLoader>
#include <KIconEffect>
#include <NETRootInfo>
#include <KWindowInfo>
#include <KWindowSystem>

#include <Plasma/Svg>
#include <Plasma/Theme>
#include <Plasma/Animator>

#include <qimageblitz/qimageblitz.h>

#ifdef Q_WS_X11
#include <QX11Info>

#include <X11/Xlib.h>
#include <fixx11h.h>
#endif

FancyTasksIcon::FancyTasksIcon(TaskManager::AbstractGroupableItem *abstractItem, FancyTasksLauncher *launcher, FancyTasksJob *job, FancyTasksApplet *parent) : QGraphicsWidget(parent)
{
    setSizePolicy(QSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed));

    setAcceptsHoverEvents(true);

    setAcceptDrops(true);

    m_abstractItem = abstractItem;

    m_launcher = launcher;

    if (job)
    {
        addJob(job);
    }

    m_applet = parent;

    m_animation = 0;

    m_animationProgress = 0;

    m_jobsProgress = 0;

    m_itemsAmount = 0;

    m_demandsAttention = false;

    m_toolTipValid = false;

    m_pressed = false;

    m_pixmapValid = false;

    m_visible = true;

    m_contextMenuVisible = false;

    m_jobsRunning = false;

    m_toolTip = NULL;

    m_color = QColor();

    m_factor = m_applet->initialFactor();

    m_animationRepeater = new QTimer(this);

    m_toolTipTimer = new QTimer(this);

    m_highlightTimer = new QTimer(this);

    m_dragTimer = new QTimer(this);

    if (m_abstractItem)
    {
        setTask(m_abstractItem);
    }
    else if (m_launcher)
    {
        m_itemType = Launcher;
    }
    else if (m_jobs.count())
    {
        m_itemType = Job;

        m_title = job->title();

        m_description = job->description();

        m_icon = job->icon();

        hide();
    }
    else
    {
        m_itemType = Other;
    }

    setLauncher(launcher);

    if (m_itemType == Group)
    {
        m_color = m_group->color();
    }
    else if (m_itemType == Startup && m_applet->startupAnimation() != FancyTasksApplet::NoAnimation)
    {
        startAnimation(1000, m_applet->animationTypes()[m_applet->startupAnimation()]);
    }

    Plasma::Animator::self()->animateElement(this, Plasma::Animator::AppearAnimation);

    setObjectName("FancyTasksIcon");

    connect(this, SIGNAL(destroyed()), m_applet, SLOT(updateSize()));
    connect(this, SIGNAL(hoverMoved(QObject*, qreal)), m_applet, SLOT(itemHoverMoved(QObject*, qreal)));
    connect(this, SIGNAL(hoverLeft()), m_applet, SLOT(itemHoverLeft()));
    connect(m_applet, SIGNAL(sizeChanged(qreal)), this, SLOT(setSize(qreal)));
    connect(m_toolTipTimer, SIGNAL(timeout()), this, SLOT(toggleToolTip()));
    connect(m_highlightTimer, SIGNAL(timeout()), this, SLOT(highlightWindow()));
}

void FancyTasksIcon::paint(QPainter *painter, const QStyleOptionGraphicsItem *option, QWidget *widget)
{
    Q_UNUSED(option)
    Q_UNUSED(widget)

    if (!m_visible)
    {
        return;
    }

    QPixmap pixmap;
    qreal size = (m_size * 0.95 * ((m_applet->moveAnimation() == FancyTasksApplet::ZoomAnimation)?m_factor:((m_applet->moveAnimation() == FancyTasksApplet::JumpAnimation)?m_applet->initialFactor():1)));

    if (m_pressed)
    {
        size *= 0.9;
    }

    if (size < 1)
    {
        return;
    }

    painter->setRenderHints(QPainter::SmoothPixmapTransform | QPainter::Antialiasing | QPainter::TextAntialiasing);

    QPixmap target = QPixmap(boundingRect().size().toSize());
    target.fill(Qt::transparent);

    QPainter targetPainter(&target);
    targetPainter.setRenderHints(QPainter::SmoothPixmapTransform | QPainter::Antialiasing | QPainter::TextAntialiasing);

    if (!m_modifiedPixmap.isNull() && (m_animation || (m_applet->moveAnimation() == FancyTasksApplet::JumpAnimation && m_animationProgress < 1)))
    {
        pixmap = m_modifiedPixmap;
    }
    else
    {
        pixmap = m_pixmap;
    }

    if (pixmap.isNull())
    {
        return;
    }

    qreal xOffset = ((m_applet->formFactor() == Plasma::Vertical && m_applet->location() == Plasma::RightEdge)?(m_size - size):0);
    qreal yOffset = (((m_applet->formFactor() != Plasma::Vertical)?(m_size - size):0) - (m_size * 0.05));

    switch (m_applet->moveAnimation())
    {
        case FancyTasksApplet::ZoomAnimation:
            if (isUnderMouse())
            {
                pixmap = KIconLoader::global()->iconEffect()->apply(pixmap, KIconLoader::Desktop, KIconLoader::ActiveState);
            }
        break;
        case FancyTasksApplet::JumpAnimation:
            if (m_animationProgress && m_animationProgress < 1)
            {
                yOffset *= ((cos(2 * PI * m_animationProgress) + 1) / 4);
            }
        break;
        case FancyTasksApplet::GlowAnimation:
            if (m_factor > 0)
            {
                QPixmap mask(size, size);
                mask.fill(m_color.isValid()?m_color:Plasma::Theme::defaultTheme()->color(Plasma::Theme::HighlightColor));

                QPixmap glow = mask;
                glow.fill(Qt::transparent);

                QPainter glowPainter(&glow);
                glowPainter.setRenderHints(QPainter::SmoothPixmapTransform | QPainter::Antialiasing);
                glowPainter.drawPixmap(QRectF((size * 0.1), (size * 0.1), (size * 0.8), (size * 0.8)), m_pixmap, QRectF(0, 0, size, size));
                glowPainter.setCompositionMode(QPainter::CompositionMode_SourceAtop);
                glowPainter.drawPixmap(0, 0, mask);
                glowPainter.end();

                QImage image = glow.toImage();

                targetPainter.setOpacity(m_factor);
                targetPainter.drawPixmap(xOffset, yOffset, QPixmap::fromImage(Blitz::blur(image, 5)));
                targetPainter.setOpacity(1);
            }

            if (isUnderMouse())
            {
                pixmap = KIconLoader::global()->iconEffect()->apply(pixmap, KIconLoader::Desktop, KIconLoader::ActiveState);
            }

            size *= 0.8;
            xOffset += (size * 0.15);
            yOffset += (size * 0.15);
        break;
        case FancyTasksApplet::SpotlightAnimation:
            if (m_factor > 0)
            {
                targetPainter.setOpacity(m_factor);
                targetPainter.drawPixmap(0, 0, m_spotlightPixmap);
                targetPainter.setOpacity(0.9);
            }
        break;
        case FancyTasksApplet::FadeAnimation:
            targetPainter.setOpacity(m_factor + 0.25);
        break;
        default:
        break;
    }

    targetPainter.drawPixmap(QRectF(xOffset, yOffset, size, size), pixmap, pixmap.rect());

    if (m_jobsRunning && m_jobs.count())
    {
        qreal rotation = 0;

        QPixmap progressPixmap = m_applet->theme()->pixmap("progress").scaled((size * 0.8), (size * 0.8), Qt::IgnoreAspectRatio, Qt::SmoothTransformation);

        targetPainter.save();
        targetPainter.translate(QPointF((xOffset + (size * 0.5)), (yOffset + (size * 0.5))));

        if (m_jobsProgress && m_jobsProgress < 100)
        {
            rotation = (m_jobsProgress?(3.6 * m_jobsProgress):(360 * m_animationProgress));

            targetPainter.rotate(rotation);
        }

        targetPainter.drawPixmap(QRectF(-(size * 0.4), -(size * 0.4), progressPixmap.width(), progressPixmap.width()), progressPixmap, progressPixmap.rect());

        targetPainter.restore();
    }

    if (m_applet->titleLabelMode() != FancyTasksApplet::NoLabel && !m_title.isEmpty() && (m_applet->titleLabelMode() == FancyTasksApplet::AlwaysShowLabel || isUnderMouse()))
    {
        QFont font = targetPainter.font();
        font.setPixelSize(size * 0.2);

        targetPainter.setFont(font);

        qreal textLength = (targetPainter.fontMetrics().width(m_title) + (3 * targetPainter.fontMetrics().width(' ')));
        qreal textFieldWidth = ((textLength > (size * 0.9))?(size * 0.9):textLength);

        QRectF textField = QRectF(QPointF(((size - textFieldWidth) / 2), ((target.height() * 0.55))), QSizeF(textFieldWidth, (size * 0.25)));

        QPainterPath textFieldPath;
        textFieldPath.addRoundedRect(textField, 3, 3);

        targetPainter.setOpacity(0.6);
        targetPainter.fillPath(textFieldPath, QBrush(Plasma::Theme::defaultTheme()->color(Plasma::Theme::BackgroundColor)));
        targetPainter.setPen(QPen(Plasma::Theme::defaultTheme()->color(Plasma::Theme::BackgroundColor).darker()));
        targetPainter.drawRoundedRect(textField, 3, 3);
        targetPainter.setPen(QPen(Plasma::Theme::defaultTheme()->color(Plasma::Theme::TextColor)));
        targetPainter.setOpacity(0.9);
        targetPainter.drawText(textField, ((textLength > textFieldWidth)?((QApplication::layoutDirection() == Qt::LeftToRight)?Qt::AlignLeft:Qt::AlignRight):Qt::AlignCenter), ((textLength > textFieldWidth)?(' ' + m_title):m_title));

        if (textLength > textFieldWidth)
        {
            QLinearGradient alphaGradient(0, 0, 1, 0);
            alphaGradient.setCoordinateMode(QGradient::ObjectBoundingMode);

            if (QApplication::layoutDirection() == Qt::LeftToRight)
            {
                alphaGradient.setColorAt(0, QColor(0, 0, 0, 255));
                alphaGradient.setColorAt(0.8, QColor(0, 0, 0, 255));
                alphaGradient.setColorAt(1, QColor(0, 0, 0, 25));
            }
            else
            {
                alphaGradient.setColorAt(0, QColor(0, 0, 0, 25));
                alphaGradient.setColorAt(0.2, QColor(0, 0, 0, 255));
                alphaGradient.setColorAt(1, QColor(0, 0, 0, 255));
            }

            targetPainter.setCompositionMode(QPainter::CompositionMode_DestinationIn);
            targetPainter.fillPath(textFieldPath, alphaGradient);
            targetPainter.setCompositionMode(QPainter::CompositionMode_SourceOver);
        }
    }

    QPixmap reflection = target.copy(0, (target.height() * 0.5), target.width(), (target.height() * 0.25));
    reflection = reflection.transformed(QTransform(1, 0, 0, 0, -1, 0, 0, 0, 1), Qt::SmoothTransformation);

    QLinearGradient gradient(QPointF(0, 0), QPointF(0, reflection.height()));
    gradient.setColorAt(0, QColor(0, 0, 0, 150));
    gradient.setColorAt(0.3, QColor(0, 0, 0, 70));
    gradient.setColorAt(0.9, Qt::transparent);

    QPainter reflectionPainter(&reflection);
    reflectionPainter.setCompositionMode(QPainter::CompositionMode_DestinationIn);
    reflectionPainter.fillRect(0, 0, reflection.width(), reflection.height(), gradient);
    reflectionPainter.end();

    targetPainter.drawPixmap(QRectF(0, ((yOffset + size) * 0.9), reflection.width(), reflection.height()), reflection, QRectF(0, 0, reflection.width(), reflection.height()));

    if (m_itemsAmount)
    {
        int amount = m_itemsAmount;

        if (amount > 5)
        {
            amount = 5;
        }

        qreal diameter = (m_size * 0.2);
        qreal xPosition = (((boundingRect().width() + xOffset) - (amount * diameter)) / 2);
        qreal yPosition = ((yOffset + size) * 1.05);

        targetPainter.setOpacity(1);

        for (int i = 0; i < amount; ++i)
        {
            targetPainter.drawPixmap(xPosition, yPosition, m_taskMarkPixmap);

            xPosition += diameter;
        }
    }

    painter->drawPixmap(0, 0, target);
}

void FancyTasksIcon::hoverEnterEvent(QGraphicsSceneHoverEvent *event)
{
    Q_UNUSED(event)

    update();

    m_toolTipTimer->start(1000);

    m_highlightTimer->start(500);

    if (m_applet->moveAnimation() == FancyTasksApplet::JumpAnimation)
    {
        Plasma::Animator::self()->customAnimation(50, 500, Plasma::Animator::LinearCurve, this, "jumpAnimation");
    }
}

void FancyTasksIcon::hoverMoveEvent(QGraphicsSceneHoverEvent *event)
{
    emit hoverMoved(this, (qreal) (((m_applet->location() == Plasma::LeftEdge || m_applet->location() == Plasma::RightEdge)?event->pos().y():event->pos().x()) / iconSize()));
}

void FancyTasksIcon::hoverLeaveEvent(QGraphicsSceneHoverEvent *event)
{
    Q_UNUSED(event)

#ifdef Q_WS_X11
    Display *dpy = QX11Info::display();
    Atom atom = XInternAtom(dpy, "_KDE_WINDOW_HIGHLIGHT", False);

    switch (m_itemType)
    {
        case Task:
            XDeleteProperty(dpy, m_task->task()->window(), atom);
        break;
        case Group:
        case Launcher:
        case Startup:
        default:
        break;
    }
#endif

    m_pressed = false;

    update();

    if (m_toolTipTimer->isActive())
    {
        m_toolTipTimer->stop();
    }
    else
    {
        m_toolTipTimer->start(200);
    }

    m_highlightTimer->stop();

    emit hoverLeft();
}

void FancyTasksIcon::dragEnterEvent(QGraphicsSceneDragDropEvent *event)
{
    if (event->mimeData()->hasFormat("FancyTasksIcon"))
    {
        event->acceptProposedAction();

        return;
    }

    m_dragTimer->stop();

    if (m_itemType == Task || m_itemType == Group)
    {
        update();

        m_dragTimer->setSingleShot(true);
        m_dragTimer->setInterval(300);
        m_dragTimer->start();

        connect(m_dragTimer, SIGNAL(timeout()), this, SLOT(activateWindow()));

        if (m_itemType == Group)
        {
            connect(m_dragTimer, SIGNAL(timeout()), this, SLOT(toggleToolTip()));
        }
    }

    if (m_itemType != Launcher)
    {
        event->ignore();
    }
}

void FancyTasksIcon::dragMoveEvent(QGraphicsSceneDragDropEvent *event)
{
    m_applet->itemDragged(this, event->pos(), event->mimeData());
}

void FancyTasksIcon::dragLeaveEvent(QGraphicsSceneDragDropEvent *event)
{
    QTimer::singleShot(500, m_applet, SLOT(hideDropZone()));

    if (m_itemType != Launcher)
    {
        event->ignore();
    }

    if (m_itemType == Group && m_toolTip && m_toolTip->isVisible())
    {
        toggleToolTip();
    }

    update();

    m_dragTimer->stop();
}

void FancyTasksIcon::dropEvent(QGraphicsSceneDragDropEvent *event)
{
    m_applet->hideDropZone();

    if (event->mimeData()->hasFormat("FancyTasksIcon") && (m_itemType == Task || m_itemType == Group))
    {
        if (m_applet->groupManager()->groupingStrategy() == TaskManager::GroupManager::ManualGrouping)
        {
            QByteArray data = event->mimeData()->data("FancyTasksIcon");
            FancyTasksIcon *droppedIcon;

            memcpy(&droppedIcon, data, sizeof(FancyTasksIcon*));

            if (m_itemType == Group && droppedIcon->itemType() == Task)
            {
                m_applet->groupManager()->manualGroupingRequest(droppedIcon->abstractItem(), m_group);
            }
            else
            {
                TaskManager::ItemList items;
                TaskManager::ItemList temp;

                items.append(m_abstractItem);

                if (m_itemType == Group)
                {
                    temp = m_group->members();

                    for (int i = 0; i < temp.count(); ++i)
                    {
                        items.append(temp.at(i));
                    }
                }

                if (droppedIcon->itemType() == Group)
                {
                    temp = droppedIcon->group()->members();

                    for (int i = 0; i < temp.count(); ++i)
                    {
                        items.append(temp.at(i));
                    }
                }
                else
                {
                    items.append(droppedIcon->abstractItem());
                }

                m_applet->groupManager()->manualGroupingRequest(items);
            }

            event->accept();
        }
        else
        {
            m_applet->dropEvent(event);

            event->ignore();
        }

        return;
    }

    if (m_itemType == Launcher)
    {
        m_launcher->dropEvent(this, event);
    }

    event->ignore();
}

void FancyTasksIcon::mousePressEvent(QGraphicsSceneMouseEvent *event)
{
    m_activeWindow = KWindowSystem::activeWindow();

    m_pressed = true;

    update();

    blockToolTip();

    if (m_demandsAttention)
    {
        stopAnimation();

        m_demandsAttention = false;
    }

    if (m_itemType != Task)
    {
        return;
    }

    publishIconGeometry();

    event->accept();
}

void FancyTasksIcon::mouseMoveEvent(QGraphicsSceneMouseEvent *event)
{
    if (QPoint(event->screenPos() - event->buttonDownScreenPos(Qt::LeftButton)).manhattanLength() < QApplication::startDragDistance() || m_itemType == Startup || (m_itemType == Launcher && m_applet->immutability() != Plasma::Mutable))
    {
        return;
    }

    blockToolTip();

    QByteArray data;
    data.resize(sizeof(FancyTasksIcon*));

    FancyTasksIcon *icon = this;

    memcpy(data.data(), &icon, sizeof(FancyTasksIcon*));

    QMimeData* mimeData = new QMimeData();
    mimeData->setData("FancyTasksIcon", data);

    if (m_itemType == Launcher)
    {
        mimeData->setData("FancyTasksLauncher", QByteArray());

        m_launcher->url().populateMimeData(mimeData);
    }
    else if (m_itemType == Task)
    {
        mimeData->setData("FancyTasksTask", QByteArray());

        m_task->addMimeData(mimeData);
    }
    else if (m_itemType == Group)
    {
        mimeData->setData("FancyTasksGroup", QByteArray());

        m_group->addMimeData(mimeData);
    }
    else
    {
        mimeData->setData("FancyTasksOther", QByteArray());
    }

    QDrag *drag = new QDrag(event->widget());
    drag->setMimeData(mimeData);
    drag->setPixmap(m_pixmap.scaled(32, 32, Qt::IgnoreAspectRatio, Qt::SmoothTransformation));
    drag->exec();
}

void FancyTasksIcon::mouseReleaseEvent(QGraphicsSceneMouseEvent *event)
{
    if (m_pressed)
    {
        m_pressed = false;

        update();
    }

    if (event->button() == Qt::LeftButton)
    {
        if (event->modifiers() & Qt::ShiftModifier)
        {
            if (m_itemType == Task)
            {
                m_task->task()->close();
            }
            else if (m_itemType == Group)
            {
                m_group->close();
            }
        }
        else
        {
            activate();
        }

        event->accept();

        return;
    }
    else if (event->button() == Qt::MidButton && m_launcher)
    {
        m_launcher->activate();

        event->accept();

        return;
    }

    event->ignore();
}

void FancyTasksIcon::contextMenuEvent(QGraphicsSceneContextMenuEvent *event)
{
    if (m_itemType == Launcher || m_itemType == Job || m_itemType == Task || m_itemType == Group)
    {
        QList<QAction*> actions;
        TaskManager::BasicMenu *taskMenu = NULL;
        KMenu *menu = new KMenu;
        menu->addTitle(m_icon, m_title.left(20));

        blockToolTip();

        if (m_itemType == Launcher)
        {
            actions = m_launcher->contextMenu()->actions();
        }
        else if (m_itemType == Job)
        {
            if (m_jobs.count() == 1)
            {
                actions = m_jobs.at(0)->contextMenu()->actions();
            }
        }
        else
        {
            if (m_itemType == Task)
            {
                taskMenu = new TaskManager::BasicMenu(NULL, m_task, m_applet->groupManager(), actions);
            }
            else
            {
                taskMenu = new TaskManager::BasicMenu(NULL, m_group, m_applet->groupManager(), actions);
            }

            if (m_launcher)
            {
                taskMenu->addSeparator();

                taskMenu->addMenu(m_launcher->contextMenu());
            }

            actions = taskMenu->actions();
        }

        m_contextMenuVisible = true;

        menu->addActions(actions);

        if (m_jobs.count() && (m_itemType != Job || m_jobs.count() > 1))
        {
            if (actions.count())
            {
                menu->addSeparator();
            }

            for (int i = 0; i < m_jobs.count(); ++i)
            {
                QAction *action = menu->addAction(m_jobs.at(i)->icon(), m_jobs.at(i)->title());
                action->setMenu(m_jobs.at(i)->contextMenu());
            }
        }

        menu->exec(m_applet->popupPosition(menu->sizeHint(), iconGeometry()));

        delete menu;

        if (taskMenu)
        {
            delete taskMenu;
        }

        m_contextMenuVisible = false;
    }

    event->accept();
}

void FancyTasksIcon::setFactor(qreal factor)
{
    if (factor == m_factor)
    {
        return;
    }

    m_factor = factor;

    if (!m_visible)
    {
        return;
    }

    if (!m_animation)
    {
        refreshPixmap(false);
    }

    updateSize();

    update();
}

void FancyTasksIcon::setSize(qreal size)
{
    size *= 0.8;

    if (!m_visible)
    {
        m_size = size;
    }

    if (size == m_size)
    {
        return;
    }

    if (m_itemsAmount)
    {
        m_taskMarkPixmap = m_applet->theme()->pixmap("mark").scaled((size * 0.15), (size * 0.15), Qt::IgnoreAspectRatio, Qt::SmoothTransformation);
    }

    if (m_applet->moveAnimation() == FancyTasksApplet::SpotlightAnimation || m_applet->demandsAttentionAnimation() == FancyTasksApplet::SpotlightAnimation || m_applet->startupAnimation() == FancyTasksApplet::SpotlightAnimation)
    {
        m_spotlightPixmap = m_applet->theme()->pixmap("spotlight").scaled(size, size, Qt::IgnoreAspectRatio, Qt::SmoothTransformation);
    }

    if (m_color.isValid())
    {
        setColor(m_color);
    }

    if (size > m_size || m_pixmap.isNull())
    {
        m_pixmapValid = false;
    }

    m_size = size;

    setIcon();

    setThumbnail();

    refreshPixmap(false);

    updateSize();
}

void FancyTasksIcon::setColor(QColor color)
{
    m_color = color;

    if (m_color.isValid())
    {
        KIconEffect effect;

        if (!m_taskMarkPixmap.isNull())
        {
            m_taskMarkPixmap = effect.apply(m_taskMarkPixmap, KIconEffect::Colorize, 1, m_color, true);
        }

        if (!m_spotlightPixmap.isNull())
        {
            m_spotlightPixmap = effect.apply(m_spotlightPixmap, KIconEffect::Colorize, 1, m_color, true);
        }
    }

    update();
}

void FancyTasksIcon::refreshPixmap(bool force)
{
    if ((m_pixmapValid && !force) || m_size < 1)
    {
        return;
    }

    m_pixmap = QPixmap(m_size, m_size);
    m_pixmap.fill(Qt::transparent);

    QPainter pixmapPainter(&m_pixmap);
    pixmapPainter.setRenderHints(QPainter::SmoothPixmapTransform | QPainter::Antialiasing | QPainter::TextAntialiasing);

    if (m_applet->useThumbnails() && !m_thumbnail.isNull() && m_itemType != Group)
    {
        QPixmap pixmap;

        if (m_thumbnail.width() > m_thumbnail.height())
        {
            pixmap = m_thumbnail.scaledToWidth(m_size, Qt::SmoothTransformation);
        }
        else
        {
            pixmap = m_thumbnail.scaledToHeight(m_size, Qt::SmoothTransformation);
        }

        pixmapPainter.drawPixmap(((m_size - pixmap.width()) / 2), ((m_size - pixmap.height()) / 2), pixmap);
    }
    else
    {
        pixmapPainter.drawPixmap(0, 0, iconPixmap(m_size));
    }

    if (m_applet->useThumbnails() && !m_thumbnail.isNull())
    {
        QPixmap icon = iconPixmap(m_size * 0.3);

        pixmapPainter.drawPixmap((m_pixmap.width() - icon.width()), (m_pixmap.height() - icon.height()), icon);
    }

    pixmapPainter.end();

    m_pixmapValid = true;

    if (m_itemType != Group)
    {
        updateToolTip();
    }

    update();
}

void FancyTasksIcon::startAnimation(int duration, QString animation)
{
    m_currentAnimationDuration = duration;

    m_currentAnimation = m_applet->animationTypes().key(animation);

    m_animationRepeater->setInterval(duration);
    m_animationRepeater->start();

    repeatAnimation();

    connect(m_animationRepeater, SIGNAL(timeout()), this, SLOT(repeatAnimation()));
}

void FancyTasksIcon::stopAnimation()
{
    m_animationRepeater->stop();

    Plasma::Animator::self()->stopCustomAnimation(m_animation);

    setOpacity(1);

    m_animation = 0;
}

void FancyTasksIcon::repeatAnimation()
{
    if (m_animation)
    {
        Plasma::Animator::self()->stopCustomAnimation(m_animation);
    }

    m_animation = Plasma::Animator::self()->customAnimation((m_currentAnimationDuration / 100), m_currentAnimationDuration, Plasma::Animator::LinearCurve, this, m_applet->animationTypes()[m_currentAnimation].toAscii());
}

void FancyTasksIcon::zoomAnimation(qreal progress)
{
    QPixmap pixmap = QPixmap(m_size, m_size);
    pixmap.fill(Qt::transparent);

    qreal size = ((m_size * 0.75) + (m_size * ((cos(2 * PI * progress) + 0.5) / 8)));

    if (size > 1)
    {
        QPainter pixmapPainter(&pixmap);
        pixmapPainter.setRenderHints(QPainter::SmoothPixmapTransform | QPainter::Antialiasing);
        pixmapPainter.drawPixmap(QRectF(((m_size - size) / 2), ((m_size - size) / 2), size, size), m_pixmap, QRectF(0, 0, m_size, m_size));
        pixmapPainter.end();
    }

    m_modifiedPixmap = pixmap;

    update();
}

void FancyTasksIcon::jumpAnimation(qreal progress)
{
    QPixmap pixmap = QPixmap(m_size, m_size);
    pixmap.fill(Qt::transparent);

    qreal size = (m_size * ((progress > 0.8)?((progress > 0.95)?progress:(1.8 - progress)):1));

    if (size > 1 && m_size > 1)
    {
        QPainter pixmapPainter(&pixmap);
        pixmapPainter.setRenderHints(QPainter::SmoothPixmapTransform | QPainter::Antialiasing);
        pixmapPainter.drawPixmap(QRectF(0, (m_size - size), m_size, size), m_pixmap, QRectF(0, 0, m_size, m_size));
        pixmapPainter.end();
    }

    m_modifiedPixmap = pixmap;

    m_animationProgress = progress;

    update();
}

void FancyTasksIcon::rotateAnimation(qreal progress)
{
    QPixmap pixmap = QPixmap(m_size, m_size);
    pixmap.fill(Qt::transparent);

    qreal width = (m_size * (0.5 + (((progress < 0.5)?progress:(1 - progress)) * 0.2)));
    qreal height = (m_size * (0.5 + (((progress < 0.5)?progress:(1 - progress)) * 0.2)));

    if (width > 1 && height > 1)
    {
        QPainter pixmapPainter(&pixmap);
        pixmapPainter.setRenderHints(QPainter::SmoothPixmapTransform | QPainter::Antialiasing);
        pixmapPainter.translate((m_size / 2), (m_size / 2));
        pixmapPainter.rotate(360 * progress);
        pixmapPainter.drawPixmap(QRectF(-(width / 2), -(height / 2), width, height), m_pixmap, QRectF(0, 0, m_size, m_size));
        pixmapPainter.end();
    }

    m_modifiedPixmap = pixmap;

    update();
}

void FancyTasksIcon::bounceAnimation(qreal progress)
{
    QPixmap pixmap = QPixmap(m_size, m_size);
    pixmap.fill(Qt::transparent);

    qreal width = (m_size * ((progress < 0.5)?((progress * 0.5) + 0.5):(1 - (progress / 2))));
    qreal height = (m_size * ((progress < 0.5)?(1 - (progress / 2)):((progress * 0.5) + 0.5)));

    if (width > 1 && height > 1)
    {
        QPainter pixmapPainter(&pixmap);
        pixmapPainter.setRenderHints(QPainter::SmoothPixmapTransform | QPainter::Antialiasing);
        pixmapPainter.drawPixmap(QRectF(((m_size - width) / 2), ((m_size - height) / 2), width, height), m_pixmap, QRectF(0, 0, m_size, m_size));
        pixmapPainter.end();
    }

    m_modifiedPixmap = pixmap;

    update();
}

void FancyTasksIcon::blinkAnimation(qreal progress)
{
    setOpacity(0.2 + ((cos(2 * PI * progress) + 0.5) / 4));
}

void FancyTasksIcon::glowAnimation(qreal progress)
{
    QPixmap pixmap = QPixmap(m_size, m_size);
    pixmap.fill(Qt::transparent);

    QPixmap mask(m_size, m_size);
    mask.fill(m_color.isValid()?m_color:Plasma::Theme::defaultTheme()->color(Plasma::Theme::HighlightColor));

    QPainter glowPainter(&pixmap);
    glowPainter.setRenderHints(QPainter::SmoothPixmapTransform | QPainter::Antialiasing);
    glowPainter.drawPixmap(QRectF((m_size * 0.1), (m_size * 0.1), (m_size * 0.8), (m_size * 0.8)), m_pixmap, QRectF(0, 0, m_size, m_size));
    glowPainter.setCompositionMode(QPainter::CompositionMode_SourceAtop);
    glowPainter.drawPixmap(0, 0, mask);
    glowPainter.end();

    QImage image = pixmap.toImage();

    QPainter pixmapPainter(&pixmap);
    pixmapPainter.setRenderHints(QPainter::SmoothPixmapTransform | QPainter::Antialiasing);
    pixmapPainter.setOpacity((cos(2 * PI * progress) + 1) / 4);
    pixmapPainter.drawPixmap(0, 0, QPixmap::fromImage(Blitz::blur(image, 7)));
    pixmapPainter.setOpacity(1);
    pixmapPainter.drawPixmap(QRectF((m_size * 0.1), (m_size * 0.1), (m_size * 0.8), (m_size * 0.8)), m_pixmap, QRectF(0, 0, m_size, m_size));

    m_modifiedPixmap = pixmap;

    update();
}

void FancyTasksIcon::spotlightAnimation(qreal progress)
{
    QPixmap pixmap = QPixmap(m_size, m_size);
    pixmap.fill(Qt::transparent);

    QPixmap background = m_applet->theme()->pixmap("spotlight").scaled(m_size, m_size, Qt::IgnoreAspectRatio, Qt::SmoothTransformation);

    if (m_color.isValid())
    {
        KIconEffect effect;

        background = effect.apply(background, KIconEffect::Colorize, 1, m_color, true);
    }

    QPainter pixmapPainter(&pixmap);
    pixmapPainter.setOpacity((cos(2 * PI * progress) + 1) / 4);
    pixmapPainter.drawPixmap(0, 0, background);
    pixmapPainter.setOpacity(0.5);
    pixmapPainter.drawPixmap(0, 0, m_pixmap);
    pixmapPainter.end();

    m_modifiedPixmap = pixmap;

    update();
}

void FancyTasksIcon::show()
{
    m_visible = true;

    if (m_itemType == Other)
    {
        return;
    }

    m_pixmapValid = false;

    updateSize();

    m_applet->updateSize();
}

void FancyTasksIcon::hide()
{
    m_visible = false;

    updateSize();

    m_applet->updateSize();
}

void FancyTasksIcon::updateSize()
{
    if (!m_visible)
    {
        setPreferredSize(QSize(0, 0));

        return;
    }

    qreal size = ((m_applet->moveAnimation() == FancyTasksApplet::ZoomAnimation)?m_factor:((m_applet->moveAnimation() == FancyTasksApplet::JumpAnimation)?m_applet->initialFactor():1));
    qreal width = (m_size * ((m_applet->location() == Plasma::LeftEdge || m_applet->location() == Plasma::RightEdge)?1.4:size));
    qreal height = (m_size * ((m_applet->location() == Plasma::BottomEdge || m_applet->location() == Plasma::TopEdge || m_applet->location() == Plasma::Floating)?1.4:size));

    setPreferredSize(QSizeF(width, height));
}

void FancyTasksIcon::taskChanged(::TaskManager::TaskChanges changes)
{
    if (m_itemType != Task && m_itemType != Group)
    {
        return;
    }

    m_title = ((m_itemType == Group)?m_group->name():m_task->task()->visibleName());

    if (m_title.isEmpty())
    {
        if (m_itemType == Group)
        {
            m_title = static_cast<TaskManager::TaskItem*>(m_group->members().at(0))->task()->visibleName();

            m_group->setName(m_title);
        }
        else
        {
            m_title = i18n("Application");
        }
    }

    m_description = (m_abstractItem->isOnAllDesktops()?i18n("On all desktops"):i18nc("Which virtual desktop a window is currently on", "On %1", KWindowSystem::desktopName(m_abstractItem->desktop())));

    if (changes & TaskManager::IconChanged)
    {
        setIcon();
    }

    if (m_abstractItem->demandsAttention())
    {
        if (!m_demandsAttention)
        {
            m_applet->needsVisualFocus();
        }

        if (m_applet->demandsAttentionAnimation() != FancyTasksApplet::NoAnimation)
        {
            startAnimation(1000, m_applet->animationTypes()[m_applet->demandsAttentionAnimation()]);

            m_demandsAttention = true;
        }
    }
    else if (m_demandsAttention)
    {
        stopAnimation();

        m_demandsAttention = false;
    }

    if (changes & TaskManager::IconChanged || changes & TaskManager::NameChanged)
    {
        m_toolTipValid = false;

        if (m_toolTip && m_toolTip->isVisible())
        {
            updateToolTip();
        }
    }

    if (m_itemType == Task)
    {
        QTimer::singleShot(200, this, SLOT(setThumbnail()));
    }
}

void FancyTasksIcon::launcherChanged()
{
    if (!m_launcher)
    {
        return;
    }

    setIcon();

    m_thumbnail = m_launcher->thumbnail();

    m_title = m_launcher->title();

    m_description = m_launcher->description();

    m_toolTipValid = false;

    refreshPixmap();
}

void FancyTasksIcon::jobChanged()
{
    QString description;
    int amount = 0;
    int percentage = 0;

    m_jobsRunning = false;

    if (m_jobs.count() > 0)
    {
        for (int i = 0; i < m_jobs.count(); ++i)
        {
            if (!m_jobs.at(i))
            {
                m_jobs.removeAt(i);

                --i;

                continue;
            }

            if (m_jobs.at(i)->percentage() && !m_jobs.at(i)->isFinished() && !m_jobs.at(i)->hasError())
            {
                ++amount;

                m_jobsRunning = true;

                percentage += m_jobs.at(i)->percentage();
            }

            description.append(m_jobs.at(i)->information() + ((i < (m_jobs.count() - 1))?"<br />":QString()));
        }
    }

    if (m_itemType == Job && m_jobs.count() == 1)
    {
        description = m_jobs.at(0)->description();
    }

    if (percentage)
    {
        percentage /= amount;
    }

    m_jobsProgress = percentage;
    m_description = description;

    if (!m_jobsRunning && (m_itemType == Task || m_itemType == Group))
    {
        taskChanged(::TaskManager::EverythingChanged);
    }

    update();

    updateToolTip();
}

void FancyTasksIcon::jobNeedsAttention()
{
    if (!m_demandsAttention)
    {
        m_applet->needsVisualFocus();
    }

    if (m_applet->demandsAttentionAnimation() != FancyTasksApplet::NoAnimation)
    {
        startAnimation(1000, m_applet->animationTypes()[m_applet->demandsAttentionAnimation()]);

        m_demandsAttention = true;
    }

    jobChanged();
}

void FancyTasksIcon::gotTaskPointer()
{
    setTask(m_abstractItem);
}

void FancyTasksIcon::setIcon()
{
    switch (m_itemType)
    {
        case Launcher:
            m_icon = m_launcher->icon();
        break;
        case Startup:
            m_icon = KIcon(m_task->startup()->icon());
        break;
        case Task:
            m_icon = KIcon(m_task->task()->icon());
        break;
        case Group:
            m_icon = KIcon(m_group->icon());
        break;
        default:
            return;
    }

    refreshPixmap();
}

void FancyTasksIcon::setThumbnail()
{
    if (m_itemType != Task)
    {
        return;
    }

    if (!m_applet->useThumbnails() && !m_thumbnail.isNull())
    {
        refreshPixmap();

        return;
    }

    QPixmap thumbnail = FancyTasksApplet::windowPreview(m_task->task()->window(), ((200 > iconSize())?200:iconSize()));

    if (!thumbnail.isNull())
    {
        m_thumbnail = thumbnail;

        refreshPixmap();
    }
}

void FancyTasksIcon::setLauncher(FancyTasksLauncher *launcher)
{
    if (m_launcher && m_itemType != Launcher)
    {
        m_launcher->removeItem(this);
    }

    m_launcher = launcher;

    if (m_launcher)
    {
        if (m_itemType != Launcher)
        {
            m_launcher->addItem(this);
        }

        launcherChanged();

        if (m_itemType == Launcher)
        {
            connect(m_launcher, SIGNAL(hide()), this, SLOT(hide()));
            connect(m_launcher, SIGNAL(show()), this, SLOT(show()));
        }
        else
        {
            disconnect(m_launcher, SIGNAL(hide()), this, SLOT(hide()));
            disconnect(m_launcher, SIGNAL(show()), this, SLOT(show()));
        }

        connect(m_launcher, SIGNAL(changed()), this, SLOT(launcherChanged()));
    }
}

void FancyTasksIcon::addJob(FancyTasksJob *job)
{
    if (m_jobs.indexOf(job) > -1)
    {
        return;
    }

    m_jobs.append(job);

    if (m_itemType == Job && m_jobs.count() > 1)
    {
        m_title = i18n("%1 jobs", m_jobs.count());
    }

    jobChanged();

    connect(job, SIGNAL(changed()), this, SLOT(jobChanged()));
    connect(job, SIGNAL(needsAttention()), this, SLOT(jobNeedsAttention()));
    connect(job, SIGNAL(close(FancyTasksJob*)), this, SLOT(removeJob(FancyTasksJob*)));
}

void FancyTasksIcon::removeJob(FancyTasksJob *job)
{
    m_jobs.removeAll(job);

    if (m_itemType == Job)
    {
        if (m_jobs.count() > 1)
        {
            m_title = i18n("%1 jobs", m_jobs.count());
        }
        else if (m_jobs.count() == 1)
        {
            m_title = m_jobs.at(0)->title();
        }
        else
        {
            deleteLater();

            return;
        }
    }

    jobChanged();
}

void FancyTasksIcon::setTask(TaskManager::AbstractGroupableItem *abstractItem)
{
    m_abstractItem = abstractItem;

    if (!m_abstractItem)
    {
        return;
    }

    if (m_abstractItem->isGroupItem())
    {
        m_group = static_cast<TaskManager::TaskGroup*>(m_abstractItem);

        m_itemType = Group;

        m_itemsAmount = m_group->members().count();

        connect(m_group, SIGNAL(changed(::TaskManager::TaskChanges)), this, SLOT(taskChanged(::TaskManager::TaskChanges)));
        connect(m_group, SIGNAL(groupEditRequest()), this, SLOT(showGroupDialog()));
        connect(m_group, SIGNAL(itemAdded(AbstractGroupableItem*)), this, SLOT(groupItemAdded(AbstractGroupableItem*)));
        connect(m_group, SIGNAL(itemRemoved(AbstractGroupableItem*)), this, SLOT(groupItemRemoved(AbstractGroupableItem*)));
    }
    else
    {
        m_task = static_cast<TaskManager::TaskItem*>(m_abstractItem);

        m_itemType = (m_task->task().isNull()?Startup:Task);

        m_itemsAmount = 1;

        connect(m_task, SIGNAL(changed(::TaskManager::TaskChanges)), this, SLOT(taskChanged(::TaskManager::TaskChanges)));
    }

    if (m_itemType == Startup)
    {
        m_title = m_task->startup()->text();

        m_description = i18n("Starting application...");

        connect(m_task, SIGNAL(gotTaskPointer()), this, SLOT(gotTaskPointer()));
    }
    else if (m_itemType == Task || m_itemType == Group)
    {
        stopAnimation();

        taskChanged(::TaskManager::EverythingChanged);

        if (m_itemType != Task)
        {
            return;
        }

        QTimer::singleShot(1000, this, SLOT(publishIconGeometry()));

        setLauncher(m_applet->launcherForTask(m_task));
    }

    update();
}

void FancyTasksIcon::activate()
{
    switch (m_itemType)
    {
        case Launcher:
            blockToolTip();

            m_launcher->activate();
        break;
        case Task:
            blockToolTip();

            if (!m_task->task()->isIconified() && m_activeWindow == m_task->task()->window())
            {
                m_task->task()->setIconified(true);
            }
            else if (!m_task->task()->isActive() || m_task->task()->isIconified())
            {
                m_task->task()->activate();
            }
            else if (!m_task->task()->isActive() && !m_task->task()->isOnTop())
            {
                m_task->task()->raise();
            }
            else
            {
                m_task->task()->setIconified(true);
            }
        break;
        case Group:
            if (!m_toolTip || !m_toolTip->isVisible())
            {
                toggleToolTip();
            }
        break;
        default:
        break;
    }
}

void FancyTasksIcon::activateWindow(bool force)
{
    if (m_itemType == Task && m_task->task() && (isUnderMouse() || force))
    {
        m_task->task()->activateRaiseOrIconify();
    }
}

void FancyTasksIcon::highlightWindow()
{
#ifdef Q_WS_X11
    if (m_itemType != Task)
    {
        return;
    }

    QVarLengthArray<long, 1024> data(1);
    Display *dpy = QX11Info::display();
    Atom atom = XInternAtom(dpy, "_KDE_WINDOW_HIGHLIGHT", False);

    data[0] = m_task->task()->window();

    XChangeProperty(dpy, m_task->task()->window(), atom, atom, 32, PropModeReplace, reinterpret_cast<unsigned char *>(data.data()), data.size());
#endif
}

void FancyTasksIcon::publishIconGeometry()
{
    if (m_itemType == Task)
    {
        m_task->task()->publishIconGeometry(iconGeometry());
    }
}

void FancyTasksIcon::updateToolTip()
{
    if (m_itemType == Group && m_toolTip)
    {
        if (m_toolTip->isVisible())
        {
            m_toolTip->hide(true);
        }

        delete m_toolTip;

        m_toolTip = NULL;
    }

    if (m_itemType == Other)
    {
        return;
    }

    QString progress;

    if (m_jobs.count())
    {
        if (m_jobsRunning && m_jobsProgress > 0)
        {
            progress.append(" <nobr><span style=\"color:" + Plasma::Theme::defaultTheme()->color(Plasma::Theme::HighlightColor).name() + ";\">");

            int bar = floor(m_jobsProgress * 0.2);

            for (int i = 0; i < 20; ++i)
            {
                progress.append("|");

                if (bar == i)
                {
                    progress.append("</span>");
                }
            }

            if (!progress.contains("</span>"))
            {
                progress.append("</span>");
            }

            progress.append(QChar(' ') + QString::number(m_jobsProgress) + "%</nobr>");
        }
    }

    if (m_toolTip)
    {
        m_toolTip->setText(QString("<p><b>%1</b><br>\n<small>%2</small></p>").arg(m_title + progress).arg(m_description));
    }
    else 
    {
        if (m_itemType == Job)
        {
            m_toolTip = new FancyTasksToolTip(iconPixmap(32), QString("<p><b>%1</b><br>\n<small>%2</small></p>").arg(m_title + progress).arg(m_description), 200, NULL);
        }
        else if (m_itemType == Task || m_itemType == Group)
        {
            m_toolTip = new FancyTasksToolTip(m_thumbnail, iconPixmap(32), QString("<p><b>%1</b><br>\n<small>%2</small></p>").arg(m_title + progress).arg(m_description), m_abstractItem, 200, NULL);
        }
        else
        {
            m_toolTip = new FancyTasksToolTip(m_thumbnail, iconPixmap(32), QString("<p><b>%1</b><br>\n<small>%2</small></p>").arg(m_title).arg(m_description), m_launcher, 200, NULL);
        }
    }

    m_toolTipValid = true;

    connect(this, SIGNAL(destroyed()), m_toolTip, SLOT(deleteLater()));
}

void FancyTasksIcon::updateToolTipPosition()
{
    m_toolTip->move(m_applet->popupPosition(m_toolTip->size(), iconGeometry(), true));
}

void FancyTasksIcon::toggleToolTip()
{
    if (isUnderMouse() && isVisible() && !m_contextMenuVisible)
    {
        if (!m_toolTipValid)
        {
            updateToolTip();
        }

        updateToolTipPosition();

        m_toolTip->show();
    }
    else if (m_toolTip)
    {
        m_toolTip->hide();
    }

    m_toolTipTimer->stop();
}

void FancyTasksIcon::blockToolTip()
{
    m_toolTipTimer->stop();

    if (m_toolTip && m_toolTip->isVisible())
    {
        m_toolTip->hide();
    }
}

void FancyTasksIcon::showGroupDialog()
{
    if (m_itemType == Group && m_applet->groupManager()->taskGrouper()->editableGroupProperties() & TaskManager::AbstractGroupingStrategy::Name)
    {
        QWidget *groupWidget = new QWidget;

        m_groupUi.setupUi(groupWidget);
        m_groupUi.icon->setIcon(m_icon);
        m_groupUi.name->setText(m_title);
        m_groupUi.color->setColor(m_color);

        m_groupDialog = new KDialog;
        m_groupDialog->setMainWidget(groupWidget);
        m_groupDialog->setButtons(KDialog::Cancel | KDialog::Ok);

        connect(m_groupDialog, SIGNAL(okClicked()), this, SLOT(groupDialogAccepted()));

        m_groupDialog->setWindowTitle(i18n("%1 Settings", m_title));
        m_groupDialog->show();
    }
}

void FancyTasksIcon::groupDialogAccepted()
{
    m_group->setIcon(KIcon(m_groupUi.icon->icon()));
    m_group->setName(m_groupUi.name->text());
    m_group->setColor(m_groupUi.color->color());

    setColor(m_groupUi.color->color());

    refreshPixmap();
}

void FancyTasksIcon::groupItemAdded(AbstractGroupableItem* item)
{
    Q_UNUSED(item)

    setTask(m_abstractItem);
}

void FancyTasksIcon::groupItemRemoved(AbstractGroupableItem* item)
{
    Q_UNUSED(item)

    setTask(m_abstractItem);
}

FancyTasksIcon::ItemType FancyTasksIcon::itemType() const
{
    return m_itemType;
}

QPointer<FancyTasksLauncher> FancyTasksIcon::launcher()
{
    return m_launcher;
}

QList<QPointer<FancyTasksJob> > FancyTasksIcon::jobs()
{
    return m_jobs;
}

TaskManager::AbstractGroupableItem* FancyTasksIcon::abstractItem()
{
    return m_abstractItem;
}

TaskManager::TaskGroup* FancyTasksIcon::group()
{
    return m_group;
}

QString FancyTasksIcon::title() const
{
    return m_title;
}

QPainterPath FancyTasksIcon::shape() const
{
    QPainterPath path;
    QRectF shape = boundingRect();
    qreal size = (m_size * ((m_applet->moveAnimation() == FancyTasksApplet::ZoomAnimation)?m_factor:((m_applet->moveAnimation() == FancyTasksApplet::JumpAnimation)?m_applet->initialFactor():1)));

    if (m_applet->formFactor() == Plasma::Vertical)
    {
        if (m_applet->location() == Plasma::LeftEdge)
        {
            shape.adjust(0, 0, -(m_size - size), 0);
        }
        else
        {
            shape.adjust(0, (m_size - size), 0, 0);
        }
    }
    else
    {
        shape.adjust(0, (m_size - size), 0, -(boundingRect().height() - m_size));
    }

    path.addRect(shape);

    return path;
}

QRect FancyTasksIcon::iconGeometry() const
{
    qreal size = (m_size * 0.95 * ((m_applet->moveAnimation() == FancyTasksApplet::ZoomAnimation)?m_factor:((m_applet->moveAnimation() == FancyTasksApplet::JumpAnimation)?m_applet->initialFactor():1)));

    QRect iconGeometry = QRect(QPoint(((m_applet->location() == Plasma::RightEdge)?(m_size - size):0), ((m_applet->location() == Plasma::BottomEdge || m_applet->location() == Plasma::Floating)?(m_size - size):0)), QSize(size, size));

    if (scene() && iconGeometry.isValid())
    {
        QGraphicsView *parentView = NULL;
        QGraphicsView *possibleParentView = NULL;

        foreach (QGraphicsView *view, scene()->views())
        {
            if (view->sceneRect().intersects(sceneBoundingRect()) || view->sceneRect().contains(scenePos()))
            {
                if (view->isActiveWindow())
                {
                    parentView = view;

                    break;
                }
                else
                {
                    possibleParentView = view;
                }
            }
        }

        if (!parentView)
        {
            parentView = possibleParentView;
        }

        if (parentView)
        {
            iconGeometry = parentView->mapFromScene(mapToScene(iconGeometry)).boundingRect().adjusted(0, 0, 1, 1);
            iconGeometry.moveTopLeft(parentView->mapToGlobal(iconGeometry.topLeft()));
        }
    }

    return iconGeometry;
}

QPixmap FancyTasksIcon::iconPixmap(qreal size)
{
    QPixmap pixmap = m_icon.pixmap(size, size);

    if (pixmap.width() < size)
    {
        pixmap = pixmap.scaled(size, size, Qt::IgnoreAspectRatio, Qt::SmoothTransformation);
    }

    return pixmap;
}

KIcon FancyTasksIcon::icon()
{
    return m_icon;
}

qreal FancyTasksIcon::iconSize() const
{
    return m_size;
}

qreal FancyTasksIcon::factor() const
{
    return m_factor;
}

bool FancyTasksIcon::isVisible() const
{
    return m_visible;
}

bool FancyTasksIcon::demandsAttention() const
{
    return m_demandsAttention;
}
