/***********************************************************************************
* Fancy Tasks: Plasmoid providing a fancy representation of your tasks and launchers.
* Copyright (C) 2009 Michal Dutkiewicz aka Emdek <emdeck@gmail.com>
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*
***********************************************************************************/

#include "FancyTasksLauncher.h"
#include "FancyTasksApplet.h"
#include "FancyTasksIcon.h"

#include <QDir>
#include <QFileInfo>

#include <KRun>
#include <KShell>
#include <KPropertiesDialog>

FancyTasksLauncher::FancyTasksLauncher(KUrl url, FancyTasksApplet *parent)
{
    m_url = url;

    m_applet = parent;

    m_executable = false;

    setUrl(url);
}

FancyTasksLauncher::~FancyTasksLauncher()
{
    for (int i = 0; i < m_items.count(); ++i)
    {
        FancyTasksIcon *icon = qobject_cast<FancyTasksIcon*>(m_items.at(i));

        if (icon)
        {
            icon->setLauncher(NULL);
        }
    }
}

void FancyTasksLauncher::dropEvent(FancyTasksIcon *parent, QGraphicsSceneDragDropEvent *event)
{
    if (!KUrl::List::canDecode(event->mimeData()))
    {
        event->ignore();

        return;
    }

    KUrl::List urls = KUrl::List::fromMimeData(event->mimeData());

    if (!urls.count())
    {
        event->ignore();

        return;
    }

    if (m_mimeType->is("inode/directory"))
    {
        parent->blockToolTip();

        Qt::DropAction dropAction = Qt::CopyAction;

        if ((event->modifiers() & Qt::ShiftModifier && event->modifiers() & Qt::ControlModifier) || event->modifiers() & Qt::AltModifier)
        {
            dropAction = Qt::LinkAction;
        }
        else if (event->modifiers() & Qt::ShiftModifier)
        {
            dropAction = Qt::MoveAction;
        }
        else if (event->modifiers() & Qt::ControlModifier)
        {
            dropAction = Qt::CopyAction;
        }
        else
        {
            KMenu *menu = new KMenu;
            QAction *moveAction = menu->addAction(KIcon("go-jump"), i18nc("@action:inmenu", "&Move Here\t<shortcut>%1</shortcut>", QKeySequence(Qt::ShiftModifier).toString()));
            QAction *copyAction = menu->addAction(KIcon("edit-copy"), i18nc("@action:inmenu", "&Copy Here\t<shortcut>%1</shortcut>", QKeySequence(Qt::ControlModifier).toString()));
            QAction *linkAction = menu->addAction(KIcon("insert-link"), i18nc("@action:inmenu", "&Link Here\t<shortcut>%1</shortcut>", QKeySequence(Qt::ControlModifier + Qt::ShiftModifier).toString()));

            menu->addSeparator();
            menu->addAction(KIcon("process-stop"), i18nc("@action:inmenu", "Cancel"));

            QAction *activatedAction = menu->exec(m_applet->popupPosition(menu->sizeHint(), parent->iconGeometry()));

            delete menu;

            if (activatedAction == moveAction)
            {
                dropAction = Qt::MoveAction;
            }
            else if (activatedAction == copyAction)
            {
                dropAction = Qt::CopyAction;
            }
            else if (activatedAction == linkAction)
            {
                dropAction = Qt::LinkAction;
            }
            else
            {
                return;
            }
        }

        switch (dropAction)
        {
            case Qt::MoveAction:
                KIO::move(urls, m_realUrl);
            break;
            case Qt::CopyAction:
                KIO::copy(urls, m_realUrl);
            break;
            case Qt::LinkAction:
                KIO::link(urls, m_realUrl);
            break;
            default:
                return;
        }
    }
    else if (m_url.isLocalFile() && (m_mimeType->is("application/x-executable") || m_mimeType->is("application/x-shellscript") || KDesktopFile::isDesktopFile(m_url.toLocalFile())))
    {
        QString arguments;
        QString command;

        for (int i = 0; i < urls.count(); ++i)
        {
            arguments += ' ' + KShell::quoteArg(urls[i].isLocalFile()?urls[i].path():urls[i].prettyUrl());
        }

        if (KDesktopFile::isDesktopFile(m_url.toLocalFile()))
        {
            KDesktopFile desktopFile(m_url.toLocalFile());
            KConfigGroup config = desktopFile.desktopGroup();

            command = config.readPathEntry("Exec", QString());

            if (command.isEmpty())
            {
                command = KShell::quoteArg(m_url.path());
            }
        }
        else
        {
            command = KShell::quoteArg(m_url.path());
        }

        KRun::runCommand(command + ' ' + arguments, NULL);
    }
}

void FancyTasksLauncher::setUrl(KUrl url)
{
    m_realUrl = m_url = url;

    m_mimeType = KMimeType::findByUrl(m_url);

    if (m_url.isLocalFile() && KDesktopFile::isDesktopFile(m_url.toLocalFile()))
    {
        KDesktopFile desktopFile(m_url.toLocalFile());
        KConfigGroup config = desktopFile.desktopGroup();

        m_exec = config.readPathEntry("Exec", QString());

        m_title = (desktopFile.readName().isEmpty()?m_url.fileName():desktopFile.readName());

        m_description = (desktopFile.readGenericName().isEmpty()?(desktopFile.readComment().isEmpty()?m_url.path():desktopFile.readComment()):desktopFile.readGenericName());

        m_icon = KIcon(desktopFile.readIcon());

        if (m_exec.isEmpty())
        {
            QString filePath = desktopFile.readUrl();

            if (filePath.isEmpty())
            {
                filePath = desktopFile.readPath();
            }

            m_realUrl = KUrl(filePath);

            m_mimeType = KMimeType::findByUrl(KUrl(m_realUrl));
        }
    }
    else
    {
        m_title = m_url.fileName();

        m_description = m_url.path();

        m_icon = KIcon(KMimeType::iconNameForUrl(url));

        if (m_title.isEmpty())
        {
            if (m_url.isLocalFile())
            {
                m_title = m_url.directory();
            }
            else
            {
                m_title = m_url.protocol();
            }
        }
    }

    if (!KDesktopFile::isDesktopFile(m_url.toLocalFile()))
    {
        KFileItemList items;
        items.append(KFileItem(m_url, m_mimeType->name(), KFileItem::Unknown));

        qreal size = ((m_applet->itemSize() > 200)?m_applet->itemSize():200);

        KIO::PreviewJob *job = KIO::filePreview(items, size, size, 0, 0, true, true, new QStringList());

        connect(job, SIGNAL(gotPreview(const KFileItem&, const QPixmap&)), this, SLOT(setThumbnail(const KFileItem&, const QPixmap&)));
    }

    m_executable = (m_url.isLocalFile() && (m_mimeType->is("application/x-executable") || m_mimeType->is("application/x-shellscript") || KDesktopFile::isDesktopFile(m_url.toLocalFile())));

    emit changed();
}

void FancyTasksLauncher::setThumbnail(const KFileItem item, const QPixmap thumbnail)
{
    Q_UNUSED(item)

    m_thumbnail = thumbnail;

    emit changed();
}

void FancyTasksLauncher::setBrowseMenu()
{
    KMenu *menu = qobject_cast<KMenu*>(sender());

    if (menu->actions().count() > 2)
    {
        return;
    }

    QString url = menu->actions()[0]->data().toString();
    QDir dir(url);
    const QStringList entries = dir.entryList(QDir::AllEntries | QDir::NoDotAndDotDot, QDir::LocaleAware | QDir::DirsFirst);

    foreach (const QString& entry, entries)
    {
        QString path = url;

        if (!path.endsWith('/'))
        {
            path.append('/');
        }

        path += entry;

        QString iconName = KMimeType::iconNameForUrl(KUrl(path));

        if (!QFileInfo(path).isFile())
        {
            KMenu *subMenu = new KMenu;

            QAction *action = subMenu->addAction(KIcon("document-open"), i18n("Open"));
            action->setData(path);

            subMenu->addSeparator();

            action = menu->addAction(KIcon(iconName), entry);
            action->setMenu(subMenu);

            connect(subMenu, SIGNAL(aboutToShow()), this, SLOT(setBrowseMenu()));
        }
        else
        {
            QAction *action = menu->addAction(KIcon(iconName), entry);
            action->setData(path);
        }
    }
}

void FancyTasksLauncher::openUrl(QAction *action)
{
    if (!action->data().isNull())
    {
       new KRun(KUrl(action->data().toString()), NULL);
    }
}

void FancyTasksLauncher::activate()
{
    new KRun(m_url, NULL);
}

void FancyTasksLauncher::showPropertiesDialog()
{
    KPropertiesDialog *propertiesDialog = new KPropertiesDialog(m_url);
    propertiesDialog->setWindowModality(Qt::NonModal);
    propertiesDialog->setWindowTitle(i18n("%1 Settings", m_title));
    propertiesDialog->show();

    connect(propertiesDialog, SIGNAL(saveAs(const KUrl&, KUrl&)), m_applet, SLOT(urlChanged(const KUrl&, KUrl&)));
}

void FancyTasksLauncher::addItem(QObject *object)
{
    if (!m_items.contains(object))
    {
        m_items.append(object);

        connect(object, SIGNAL(destroyed(QObject*)), this, SLOT(removeItem(QObject*)));
    }

    if (m_items.count() == 1)
    {
        emit hide();
    }
}

void FancyTasksLauncher::removeItem(QObject *object)
{
    m_items.removeAll(object);

    if (!m_items.count())
    {
        emit show();
    }
}

int FancyTasksLauncher::itemsAmount() const
{
    return m_items.count();
}

bool FancyTasksLauncher::isExecutable() const
{
    return m_executable;
}

KMenu* FancyTasksLauncher::contextMenu()
{
    KMenu *menu = new KMenu;
    QAction *openAction = menu->addAction(KIcon("system-run"), i18n("Run"));

    connect(openAction, SIGNAL(triggered()), this, SLOT(activate()));

    if (m_mimeType->is("inode/directory"))
    {
        KMenu *browseMenu = new KMenu;

        QAction *action = browseMenu->addAction(KIcon("document-open"), i18n("Open"));
        action->setData(m_url.path());

        browseMenu->addSeparator();

        action = menu->addAction(KIcon("document-preview"), i18n("Browse"));
        action->setMenu(browseMenu);

        connect(browseMenu, SIGNAL(aboutToShow()), this, SLOT(setBrowseMenu()));
        connect(browseMenu, SIGNAL(triggered(QAction*)), this, SLOT(openUrl(QAction*)));
    }

    QAction *editAction = menu->addAction(KIcon("document-edit"), i18n("Edit"));

    connect(editAction, SIGNAL(triggered()), this, SLOT(showPropertiesDialog()));

    menu->setTitle(i18n("Launcher"));

    return menu;
}

KIcon FancyTasksLauncher::icon()
{
    return m_icon;
}

KUrl FancyTasksLauncher::url() const
{
    return m_url;
}

QPixmap FancyTasksLauncher::thumbnail()
{
    return m_thumbnail;
}

QString FancyTasksLauncher::exec() const
{
    return m_exec;
}

QString FancyTasksLauncher::title() const
{
    return m_title;
}

QString FancyTasksLauncher::description() const
{
    return m_description;
}
